import copy
from typing import Dict,List, Tuple, Optional
import multiprocessing

from tqdm import tqdm
import numpy as np

from qrem.functions_qrem import functions_data_analysis as fda, ancillary_functions as anf

from qrem.common.printer import qprint
from qrem.common import convert


def is_valid_probability_vector(examined_vector: List[float], threshold=1e-5) -> bool:
    """Checks if given vector is valid probability vector i_index.e. has only positive values that sums to 1.

    Parameters
    ----------
    examined_vector: List[float]
        Vector of probabilities for which the validity is checked.
    threshold: float
        Error cluster_threshold when determining probabilities sum condition.

    Returns
    -------
        Bool information whether examined_vector is valid probability vector or not.
    """
    values_sum = 0

    for value in examined_vector:
        if value < 0:
            return False
        values_sum += value

    return abs(values_sum - 1) < threshold


def random_stochastic_matrix(size, type='left',diagonal = 1.0):
    """Get a random stochastic matrix of given type ( and size."""
    matrix = np.random.rand(size, size)+diagonal*np.identity(size)

    if (type == 'left'):
        matrix = matrix / matrix.sum(axis=0)[None, :]
    elif (type == 'right'):
        matrix = matrix / matrix.sum(axis=1)[:, None]
    elif (type == 'doubly' or type == 'double' or type == 'orto'):
        matrix = matrix = matrix / matrix.sum(axis=0)[None, :]
        matrix = matrix / matrix.sum(axis=1)[:, None]

    return matrix

# ===================================================
# Basic function used to compute marginals from experimental counts - probability distributions
# ===================================================

#JT: Function used to compute marginals
#MOcomm: this function doubles functionalities of some of 
def _compute_marginals(
        results_dictionary: Dict[str, Dict[str, int]],
        subsets_list: List[Tuple],
        print_progress_bar : bool =False,
        normalization : bool = False
):

    items_iterable = results_dictionary.items()

    if print_progress_bar:
        items_iterable = tqdm(items_iterable)

    marginals_dictionary = {}

    #JT: this for loop goes over experimental settings and results of experiments for a given setting

    for experiment_label, experimental_results in items_iterable:

               #JT: copy creates a shallow copy of an object (changes in original object influence copy)

        #subsets_list_temp = subsets_list[experiment_label]

        if isinstance(subsets_list, dict):
            subsets_list = subsets_list[experiment_label]
        
           #JT: the variable below stores unique lengths of qubits subsets

        unique_lengths = np.unique([len(x) for x in subsets_list])

        #JT: a nested dictionary
        #length -elements of unique_lengths list
        #index outcome - goes over all bit strings of length from unique_lengths list
        #index outcome is converted to a bitstring, if this is shorter than length it is filled with 0 to the left, this is key of the second dictionary
        #the value of the second dictionary is set to 0

        local_registers_dictionaries = {length:
                                            {convert.integer_to_bitstring(index_outcome,
                                                                    length):
                                                0.0
                                            for index_outcome in range(int(2 ** length))}
                                        for length in unique_lengths}
        
        vectors_dictionaries = {subset: copy.copy(local_registers_dictionaries[len(subset)]) for
                                subset in subsets_list}

        # vectors_list = [copy.copy(local_registers_dictionaries[len(subset)]) for
        #                         subset in subsets_list]
        # norm = 0


   

        #JT: this is a loop over different subsets

        for subset_index in range(len(subsets_list)):

            #JT: consecutive subsets are chosen

            subset = subsets_list[subset_index]

            #JT: For a fixed experimental setting a loop over different

            for outcome_bitstring, number_of_occurrences in experimental_results.items():

                #JT: outcome bitstring for a given subset is established

                tuple_now = ''.join([outcome_bitstring[qi] for qi in subset])

                #JT: otcome statistics for this subset and outcome string are added

                vectors_dictionaries[subset][tuple_now] += number_of_occurrences

        # vectors_dictionaries = {subsets_list[subset_index]:vectors_list[subset_index]
        #                         for subset_index in range(len(subsets_list))}


            # norm += number_of_occurrences

       #JT the dictionary has a structure key: subset value, value an array with number of occurrences
       #JT: I'm not sure, how the ordering of results id done, i.e. what is the relation between a place in array and the resulting bitstring
       #JT: I assume that 0 etry is all 0 result and so on, but this needs to be double-checked

        marginals_dict_for_this_experiment = {
            subset: np.array(list(vectors_dictionaries[subset].values())) for subset in
            subsets_list}
        marginals_dictionary[experiment_label] = marginals_dict_for_this_experiment

    return marginals_dictionary


# ===================================================
# This function is used in marginals_analyzer_base
# ===================================================

def compute_marginals(results_dictionary: Dict[str, Dict[str, int]],
        subsets_list:List[Tuple],
        use_multiprocessing : bool =False,
        number_of_threads=None,

        print_progress_bar : bool =False,
        normalization : bool = False
                                   ):
    if use_multiprocessing:
        if number_of_threads is None:
            number_of_threads = multiprocessing.cpu_count()-1

        experimental_keys = list(results_dictionary.keys())
        number_of_experiments = len(experimental_keys)

        batch_size = -int(number_of_experiments//(-number_of_threads))

        arguments_list = []
        for thread_index in range(number_of_threads):
            slice_indices_now = slice(thread_index * batch_size,
                            (thread_index + 1) * batch_size)

            print_progress_bar = False

            keys_slice_now = experimental_keys[slice_indices_now]


            #that's the last pool thread with maximal number of function calls
            # (last thread has the least of them)
            if thread_index in[number_of_threads-2]:
                print_progress_bar=True

            dictionary_res_now = {key:results_dictionary[key] for key in keys_slice_now}

            arguments_list.append((dictionary_res_now,
                                   subsets_list,
                                   print_progress_bar))
     


        pool = multiprocessing.Pool(number_of_threads)
        results = pool.starmap_async(_compute_marginals,
                                 arguments_list
                                 )
    

        pool.close()
        qprint("\nJoining pool...")
        pool.join()
        qprint("\nGetting results from pool...")

        res_multiprocessing = results.get()



        all_results = {}

        for dictionary_thread in res_multiprocessing:
            all_results = {**all_results, **dictionary_thread}

        return all_results
    else:
        return _compute_marginals(results_dictionary=results_dictionary,subsets_list=subsets_list)
# ===================================================
# Functions used to marginalize probability distributions
# ===================================================

def compute_marginal_of_probability_distribution(
        global_probability_distribution: np.ndarray,
        bits_of_interest: List[int],
        register_names: Optional[List[str]] = None) -> np.ndarray:
    """Return marginal distribution from vector of global distribution
    :param global_probability_distribution: distribution on all bits
    :param bits_of_interest: bits we are interested in (so we average over other bits)
                            Assuming that qubits are labeled
                            from 0 to log2(len(global_probability_distribution))
    :param register_names: bitstrings register, default is
                           '00...00', '000...01', '000...10', ..., etc.

    :return: marginal_distribution : marginal probability distribution

    NOTE: we identify bits with qubits in the variables bitstring_names

    #TODO FBM: do some speed tests on some details of those solutions
    """

    if len(bits_of_interest) == 0:
        print('0 length bits list')
        return global_probability_distribution

    try:
        if isinstance(global_probability_distribution[0][0], complex) or isinstance(
                global_probability_distribution[0][0], np.complex128):
            global_probability_distribution = global_probability_distribution.real
    except(IndexError):
        if isinstance(global_probability_distribution[0], complex) or isinstance(
                global_probability_distribution[0], np.complex128):
            global_probability_distribution = global_probability_distribution.real

    global_dimension = len(global_probability_distribution)
    global_number_of_qubits = int(np.log2(global_dimension))
    all_qubits = list(range(global_number_of_qubits))
    bits_to_average_over = list(set(all_qubits).difference(set(bits_of_interest)))

    number_of_bits_in_marginal = global_number_of_qubits - len(bits_to_average_over)
    dimension_of_marginal = 2 ** number_of_bits_in_marginal

    if register_names is None:
        bitstring_names = anf.get_classical_register_bitstrings(list(range(global_number_of_qubits)),
                                                                global_number_of_qubits)
    else:
        bitstring_names = register_names

    marginal_distribution = np.zeros((dimension_of_marginal, 1), dtype=float)
    for j in range(global_dimension):
        # this is slightly faster than going through "for bitstring_global in bitstring_names
        # and then converting bitstring_global to integer
        # and also faster than creating the global bitstring in situ
        bitstring_global = bitstring_names[j]

        bitstring_local = ''.join(
            [bitstring_global[qubit_index] for qubit_index in bits_of_interest])

        marginal_distribution[int(bitstring_local, 2)] += global_probability_distribution[j]

    return marginal_distribution


import itertools
import scipy
import numpy as np
import sys





def log_coefficient(number_of_symbols,locality):
    """
    A helper function needed to compute probabilistic bound on number of circuits, for which all combinations of tomographic settings are realized for all subsets of locality k
    Parameters
    ----------
    number_of_symbols: int
        number of different symbols used to construct tomographic circuits: 2 for DDOT circuits, 6 for QDOT circuits (overcomplete Pauli basis)  

    locality: int
        locality of sets for which all the symbols should appear 

    Returns
    -------

    A particular expression needed to calculate probabilistic bound on number of informationally (over)complete set of circuits

    """
    return 1/np.log(number_of_symbols**locality/(number_of_symbols**locality-1))


def circuits_number_estimation(number_of_symbols, locality, number_of_qubits,probability):
    """
    Function cacluclating a probabilitic bound on number of circuits, for which all combinations of tomographic settings are realized for all subsets of locality k
    Parameters
    ----------
    number_of_symbols: int
        number of different symbols used to construct tomographic circuits: 2 for DDOT circuits, 6 for QDOT circuits (overcomplete Pauli basis)  

    locality: int
        locality of sets for which all tomographic settings should appear

    number_of_qubits: 
        number of qubits for which tomographic circuits are realizeed

    probability:
        accepted probability of bound failiure 

    Returns
    -------

    number of circuits needed to realize all combinations of tomographic settings for all subsets of given locality
    """
    return log_coefficient(number_of_symbols,locality)*(np.log( scipy.special.binom(number_of_qubits, locality) ) +locality *np.log(number_of_symbols)+np.log(1./probability))





def information_completeness_verification_quantitative(tomographic_circuits_list, symbol_string_list, locality=2, subsets_list=None):

    """
    Function checking how many times a particular input setting to measurement tomography protocol is realized in a
    given set of circuits for all subsets of qubits of a fixed locality


    Parameters
    ----------
    tomographic_circuits_list : list
        the list of circuits that are used for measurement tomography protocol 

    locality : int
        number corresponding to locality of subsets for which information completness check is performed  

    symbol_string : string
    list of strings of symbols used to encode input settings, e.g. for 6 state overcomplete Pauli basis it's ['0','1','2','3','4','5']

      subsets_list : list of tuples
        list of qubits subsets for which check is performed, by defult set to None

    Returns
    -------
    symbols_subset_dictionary: dictionary a dictionary with keys of a form ((subset_of_qubits), input_setting)
    (e.g. ((0,1),'50')  and values corresponding to the number of times that a given input_setting is
    realized for this particular subset

    missing_symbols: dictionary a dictionary with keys corresponding to qubit_subsets (e.g. (0,1)) and values
    corresponding to list of settings that don't appear for a given subset e.g. ['00','15']
    """
    number_of_qubits = len(tomographic_circuits_list[0])

    #subsets_of_qubits = []
    #symbols_subset_dictionary = {}
    # a list of relevant marginals is created
    #for element in itertools.combinations(range(number_of_qubits), locality):
    #    subsets_of_qubits.append(element)

    symbols_subset_dictionary ={}


    #if no subsets list was provided the check is performed over all subsets of given locality
    if subsets_list==None:
        subsets_of_qubits = []
        symbols_subset_dictionary = {}
        
        # a list of relevant marginals is created
        #for element in itertools.combinations(range(number_of_qubits), locality):
        #    subsets_of_qubits.append(element)
            # a list of all combinations of relevant tomographic symbols in crated
        #symbol_list = itertools.product(symbol_string_list, repeat=locality)
        #tomographic_symbols_list = []
        # a list with all tomographic settings is created
        #for symbol in symbol_list:
            # a joint string corresponding to tomography setting is
        #    key_temp = ''
        #    for str_el in symbol:
        #        key_temp += str_el
        #    tomographic_symbols_list.append(key_temp)
    
    #if the subsets list was provided these subsets are used to perform check  
    else:
        subsets_of_qubits =subsets_list
    


    # this is a loop over all marginals
    for element in subsets_of_qubits:

        locality = len(element)
        # a list of all combinations of relevant tomographic symbols in crated
        symbol_list = itertools.product(symbol_string_list, repeat=locality)
        tomographic_symbols_list = []
        # a list with all tomographic settings is created
        for symbol in symbol_list:
            # a joint string corresponding to tomography setting is
            key_temp = ''
            for str_el in symbol:
                key_temp += str_el
            tomographic_symbols_list.append(key_temp)

        # total numbel of tomographic circuits is updated

        total_number_of_symbols = len(symbol_string_list) ** locality * len(element)
             

        # an empty dictionary with all possible combinations of symbols is created
        for symbol in tomographic_symbols_list:
            # a joint string corresponding to tomography setting is

            key = tuple((element, symbol))
            # a dictionary is initialized
            symbols_subset_dictionary[key] = 0

        # an inner loop over circuits
        for experimental_setting in tomographic_circuits_list:
            key_temp = ''

            # a circuit realized for a particular symbol is constructed
            for i in range(locality):
                key_temp += experimental_setting[element[i]]

            # a key for dictionary is constructed
            key = tuple((element, key_temp))

            # value of dictionary is updated
            if key in symbols_subset_dictionary.keys():
                temporary_value = symbols_subset_dictionary[key]
                temporary_value = temporary_value + 1
                symbols_subset_dictionary[key] = temporary_value

    # here settings that do not appear for particular marginal are saved
    missing_symbols = {}
    for key, occurrence in symbols_subset_dictionary.items():
        if occurrence == 0:
            if key[0] in missing_symbols.keys():
                temporary_value = missing_symbols[key[0]]
                temporary_value.append(key[1])
                missing_symbols[key[0]] = temporary_value
            else:
                missing_symbols[key[0]] = [key[1]]

    return symbols_subset_dictionary, missing_symbols





def information_completeness_verification_qualitative(tomographic_circuits_list, locality, symbol_string_list):
    """
    Function reports fraction of missing tomography settings in a given set of circuits for all subsets of qubits of a
    fixed locality
    Parameters
    ----------
    tomographic_circuits_list : list
        the list of circuits that are used for measurement tomography protocol

    locality : int
        number corresponding to locality of subsets for which information completeness check is performed

    symbol_string_list : string
        list of strings of symbols used to encode input settings, e.g. for 6 state overcomplete Pauli basis it's
        ['0','1','2','3','4','5']

    Returns
    -------
     missing_symbols_fraction: real
        a fraction of missing symbols for given subset structure
    """
    subsets_of_qubits = []

    number_of_qubits = len(tomographic_circuits_list[0])
    # a list of relevant marginals is created
    for element in itertools.combinations(range(number_of_qubits), locality):
        subsets_of_qubits.append(element)

    # How many symbols are involved in tomography
    symbols_number = len(symbol_string_list)
    # what is the total number of symbols times the total number of subsets
    total_number_of_symbols = symbols_number ** locality * scipy.special.comb(number_of_qubits, locality)

    symbols_subset_dictionary = {}
    # an outer loop over subsets
    for element in subsets_of_qubits:

        # an inner loop over circuits
        for experimental_setting in tomographic_circuits_list:
            key_temp = ''

            # a circuit realized for a particular symbol is constructed
            for i in range(locality):
                key_temp += experimental_setting[element[i]]

            # a key for dictionary is constructed
            key = tuple((element, key_temp))

            # a dictionary entry corresponding to realized subset symbol is created
            symbols_subset_dictionary[key] = 1

    symbols_number = len(symbols_subset_dictionary.keys())

    # a fraction of missing symbols is returned
    missing_symbols_fraction = (total_number_of_symbols - symbols_number) / (total_number_of_symbols)

    return missing_symbols_fraction



import copy
import time
from typing import Optional, Dict, List, Tuple
import multiprocessing as mp

import numpy as np

from qrem.functions_qrem import functions_data_analysis as fda
from qrem.noise_characterization.base_classes.marginals_analyzer_interface import MarginalsAnalyzerInterface
from qrem.common import probability as qrem_common_prob

from qrem.common.printer import qprint


class MarginalsAnalyzerBase(MarginalsAnalyzerInterface):
    """
    This is base class for all the classes that will operate on marginal probability distributions.
    Methods of this class allow to calculate marginal distributions from experimental results.

    In this class and its children, we use the following convention for:

     1. Generic experimental results:
    :param results_dictionary: Nested dictionary with following structure:

    results_dictionary[label_of_experiment][bitstring_outcome] = number_of_occurrences

    where:
        -label_of_experiment is arbitrary label for particular experiment,
        -bitstring_outcome is label for measurement outcome,
        -number_of_occurrences is number of times that bitstring_outcome was recorded

        Hence top-level key labels particular experiment
        (one can think about quantum circuit implementation)
        and its value is another dictionary with results of given experiment in the form
        of dictionary of measurement outcomes


    2. Results represented as marginal probability distributions:
        :param marginals_dictionary: Nested dictionary with the following structure:

        marginals_dictionary[label_of_experiment][label_of_subset] = marginal_probability_vector

        where:
            -label_of_experiment is the same as in results_dictionary and it labels results from which
            marginal distributions were calculated
            -label_of_subset is a label for qubits subset for which marginals_dictionary were calculated.
            We use convention that such label label is tuple of integers indicating qubits' indices
            -marginal_probability_vector marginal distribution stored as vector

    """

    # TODO FBM: add coarse-graining functions_qrem for marginals_dictionary as class methods

    def __init__(self,
                 results_dictionary: Dict[str, Dict[str, int]],
                 bitstrings_right_to_left: Optional[bool]=False,
                 marginals_dictionary: Optional[Dict[str, Dict[Tuple[int], np.ndarray]]] = None,
                 ) -> None:

        """
        :param results_dictionary: see class description

        :param bitstrings_right_to_left: specify whether bitstrings
                                    should be read from right to left (when interpreting qubit labels)
        :param marginals_dictionary: see class description

        NOTE: when user does not provide marginals_dictionary we create it during class initialization.
        To this aim, we create "key_dependent_dicts" (see below)
        """

        if marginals_dictionary is None:
            # If user does not provide dictionary with marginals_dictionary, we create template.
            # Instead of standard dictionaries, we use ones that are "key dependent" (see description
            # of that function), which is useful for some calculations. This is because it allows to
            # not care whether given probability distribution was already created (as value in
            # dictionary) - if not, it creates it on the run.
            marginals_dictionary_key_dependent = {key: {}
                                                  for key in results_dictionary.keys()}
        else:
            marginals_dictionary_key_dependent = {}
            # print(marginals_dictionary)
            for experiment_key, local_dict in marginals_dictionary.items():

                if isinstance(local_dict, dict):
                    new_dict = {}
                    for key, value in local_dict.items():
                        new_dict[key] = value
                else:
                    new_dict = local_dict
                marginals_dictionary_key_dependent[experiment_key] = new_dict

        # print(list(results_dictionary.values())[0])
        # raise KeyError

        # set initial values of class properties
        self._results_dictionary = results_dictionary
        self._marginals_dictionary = marginals_dictionary_key_dependent
        self._bitstrings_right_to_left = bitstrings_right_to_left


        first_experiment = list(results_dictionary.values())[0]
        # print(first_experiment)
        try:
            self._number_of_qubits = len(list(first_experiment.keys())[0])
        except(AttributeError):
        #     print(results_dictionary)
            print(type(first_experiment))
        #
            raise KeyboardInterrupt


        # print(self._number_of_qubits)
        self._qubit_indices = list(range(self._number_of_qubits ))



    @property
    def results_dictionary(self) -> Dict[str, Dict[str, int]]:

        return self._results_dictionary

    @results_dictionary.setter
    def results_dictionary(self, results_dictionary: Dict[str, Dict[str, int]]) -> None:
        self._results_dictionary = results_dictionary

    @property
    def marginals_dictionary(self) -> Dict[str, Dict[Tuple[int], np.ndarray]]:

        # if np.any([str(type(dictionary))!="<class 'dict'>" for dictionary in self._marginals_dictionary.values()]):
        #     dict_to_return = {}
        #     for experiment_key, key_dependent_dict in self._marginals_dictionary.items():
        #         dict_to_return[experiment_key] = dict(key_dependent_dict)
        # else:
        #     dict_to_return = self._marginals_dictionary
        #
        #
        # if str(type(dict_to_return))!="<class 'dict'>":
        #     dict_to_return = dict(dict_to_return)
        #
        #
        # self._marginals_dictionary = dict_to_return


        return self._marginals_dictionary

    @marginals_dictionary.setter
    def marginals_dictionary(self,
                             marginals_dictionary: Dict[str, Dict[Tuple[int], np.ndarray]]) -> None:
        self._marginals_dictionary = marginals_dictionary


    def results_dictionary_update(self,
                                  results_dictionary_new: Dict[str, Dict[str, int]]) -> None:
        # This method_name updates results dictionary from class property with new dictionary.
        # Note that if there is KEY collision, then the value from new dictionary overwrites old one.

        self._results_dictionary = {**self._results_dictionary,
                                    **results_dictionary_new}

    def marginals_dictionary_update(self,
                                    marginals_dictionary_new: Dict[
                                        str, Dict[Tuple[int], np.ndarray]]) -> None:
        # See description of self.results_dictionary_update

        self._marginals_dictionary = {**self._marginals_dictionary,
                                      **marginals_dictionary_new}







    def normalize_marginals(self,
                            experiments_keys: Optional[List[str]] = None,
                            marginals_keys: Optional[Dict[str,Tuple[int]]] = None) -> None:
        """Go through marginals_dictionary stored as class' property
           and normalize marginal distributions
        :param experiments_keys: labels for experiments
        :param marginals_keys: labels for qubit subsets_list
        """
        # If no labels of experiments are provided, we take all of them
        if experiments_keys is None:
            experiments_keys = self._marginals_dictionary.keys()

        # Loop through all experiments and marginals_dictionary and normalize them.
        for key_experiment in experiments_keys:
            if marginals_keys is None:
                # if no marginal keys are provided, we take all of them
                looping_over = self._marginals_dictionary[key_experiment].keys()
            else:
                looping_over = marginals_keys[key_experiment]

            for key_marginal in looping_over:
                self._marginals_dictionary[key_experiment][key_marginal] *= 1 / np.sum(
                    self._marginals_dictionary[key_experiment][key_marginal])


    def compute_unnormalized_marginals(self,
                                       experiment_keys: List[str],
                                       subsets_list: List[Tuple[int]],
                                       # normalize_marginals=False,
                                       multiprocessing=False) -> None:
        """Return dictionary of marginal probability distributions from counts dictionary
        :param experiment_keys: list of keys that label experiments for which marginals_dictionary should be taken
        :param subsets_dictionary: list of subsets_list of qubits for which marginals_dictionary should be calculated
        """

        #TODO FBM: update class property

        if isinstance(experiment_keys,str):
            experiment_keys = [experiment_keys]

     
        multiple_experimental_results={key:self._results_dictionary[key] for key in experiment_keys}
  

        return qrem_common_prob.compute_marginals(results_dictionary=multiple_experimental_results,subsets_list=subsets_list,use_multiprocessing=multiprocessing)






    def compute_marginals_old(self,
                          experiment_keys: List[str],
                          subsets_list: List[Tuple[int]]) -> None:
        """Return dictionary of marginal probability distributions from counts dictionary
        :param experiment_keys: list of keys that label experiments for which marginals_dictionary should be taken
        :param subsets_list: list of subsets_list of qubits for which marginals_dictionary should be calculated
        """

        if isinstance(experiment_keys, str):
            experiment_keys = [experiment_keys]

        # new_subsets_to_normalize = []
        for experiment_label in experiment_keys:
            experimental_results = self._results_dictionary[experiment_label]

            for subset in subsets_list:
                # if tuple(subset) in self._marginals_dictionary[experiment_label].keys():
                #     continue
                # else:
                # new_subsets_to_normalize.append(tuple(subset))

                # initialize marginal distribution
                marginal_vector_now = np.zeros((int(2 ** len(subset)), 1),
                                               dtype=float)
                for outcome_bitstring, number_of_occurrences in experimental_results.items():
                    if self._bitstrings_right_to_left:
                        # here we change the order of bitstring if it was specified
                        outcome_bitstring = outcome_bitstring[::-1]

                    # get bitstring denoting state of qubits in the subset
                    marginal_key_now = ''.join([outcome_bitstring[b] for b in subset])

                    # add counts to the marginal distribution
                    marginal_vector_now[int(marginal_key_now, 2)] += number_of_occurrences

                # Here if there is no "qubits_string" KEY we use the fact that by default we use
                # "key_dependent_dictionary". See description of __init__.
                self._marginals_dictionary[experiment_label][tuple(subset)] += marginal_vector_now
                # print(marginal_vector_now, marginal_vector_now/np.sum(marginal_vector_now))

        # self.normalize_marginals(experiment_keys, new_subsets_to_normalize)


    #JT: this is the method to compute marginals
    #MOcomm - name of this method is a bit wierd - I suggest changing it
    def compute_all_marginals(self,
                              subsets_dictionary: List[Tuple[int]],
                              multiprocessing=False,
                              check_if_already_calculated=False,
                              show_progress_bar = True) -> None:

        """
        #TODO FBM: add multiprocessing for this task
        #TODO FBM: change subsets_list to be dictionary
        Implements self.compute_marginals for all experimental keys.

        :param subsets_dictionary: list of subsets of qubits for which marginals_dictionary should be calculated
        :param show_progress_bar: if True, shows progress bar. Requires "tqdm" package
        """


        t0 = time.time()

        #JT: key list is a list of different experiment settings

        keys_list = list(self._results_dictionary.keys())
        keys_list_range = range(len(keys_list))

        # if progress bar should be shown, we use tqdm package
        if show_progress_bar:
            from tqdm import tqdm
            keys_list_range = tqdm(keys_list_range)

        #JT: subsets_dictionary -  A dictionary with keys -experimental settings and values - lists of subsets

        if isinstance(subsets_dictionary, list):



            subsets_dict = {key:subsets_dictionary for key in keys_list}
            subsets_dictionary = subsets_dict

        # print('\nsubsets input:', subsets_list)
        # print('\nsubsets existing:',self._marginals_dictionary)

        subsets_to_compute = {key:[] for key in keys_list}

        #JT: Loop below checks wheter some of the marginals have been computed previously

        if check_if_already_calculated:

            keys_outer_computed = self._marginals_dictionary.keys()
            for key in keys_list:

                if key not in keys_outer_computed:
                    subsets_to_compute[key] = subsets_dictionary[key]
                    # print(key)
                    # raise KeyboardInterrupt("WTF")
                else:
                    keys_inner_computed = self._marginals_dictionary[key].keys()
                    # for subset in subsets_dictionary:
                    #     if subset not in keys_inner_computed:
                    #         subsets_to_compute[key].append(subset)
                    subsets_to_compute[key] = [subset
                                              for subset in subsets_dictionary[key]
                                              if subset not in
                                              keys_inner_computed]
                    # print('hej2')
                    # print(keys_inner_computed)

        else:
            subsets_to_compute = {key:subsets_dictionary[key] for key in keys_list}

        #
        # print('\nsubsets to compute:', subsets_to_compute)
        # raise KeyboardInterrupt

        if not multiprocessing:
            big_marginals_dict = {}

            #This is a loop that goes over consecutive settings of tomographic experiments

            for key_index in keys_list_range:
                key = keys_list[key_index]

               #mrginals computation for a fixed setting is performed

                big_marginals_dict[key] = self.compute_unnormalized_marginals([key],
                                                                              #{key:subsets_to_compute[key]},
                                                                              subsets_to_compute[key],
                                                                              multiprocessing=False)[key]

        else:

            big_marginals_dict = self.compute_unnormalized_marginals(experiment_keys=keys_list,
                                                                     subsets_list=subsets_to_compute,
                                                                     multiprocessing=True)
        # print(big_marginals_dict)
        # raise KeyboardInterrupt
        t1=time.time()
        for experiment_key, marg_dict in big_marginals_dict.items():

            # print(marg_dict)
            if experiment_key not in self._marginals_dictionary.keys():
                self._marginals_dictionary[experiment_key] = {}

            for subset, marginals_dictionary_local in marg_dict.items():

                # if subset not in self._marginals_dictionary[experiment_key].keys():
                # TODO FBM: perhaps consider addition instead of replacement
                self._marginals_dictionary[experiment_key][subset] = marginals_dictionary_local
                # else:
                #
                #     print(subset, marginals_dictionary_local)
                #     self._marginals_dictionary[experiment_key][subset] = {**self._marginals_dictionary[experiment_key][subset],
                #                                                           **marginals_dictionary_local}


        t2 = time.time()

        self.normalize_marginals(experiments_keys=keys_list,
                                 marginals_keys=subsets_to_compute)

        t3=time.time()
        #
        #qprint("\nCalculating marginals took:",t1-t0)
        #qprint("Updating dictionary in class took:",t2-t1)
        #qprint("Normalizing marginals took:",t3-t2)
        #
        #




    def get_marginals(self,
                      experiment_key: str,
                      subsets_list: List[Tuple[int]]) -> Dict[Tuple[int], np.ndarray]:
        """Like self.compute_marginals but first checks if the marginals_dictionary are already computed
            and it returns them.

        :param experiment_key: key that labels experiment from which marginals_dictionary should be taken
        :param subsets_list: list of subsets_list of qubits for which marginals_dictionary should be calculated

        :return: marginals_dictionary:
                dictionary in which KEY is label for subset, and VALUE is marginal on that subset
        """


        for subset in subsets_list:
            if experiment_key not in self._marginals_dictionary.keys():
                self.compute_unnormalized_marginals([experiment_key], [subset])
            elif subset not in self._marginals_dictionary[experiment_key].keys():
                self.compute_unnormalized_marginals([experiment_key], [subset])

        return {subset: self._marginals_dictionary[experiment_key][subset]
                for subset in subsets_list}

    @staticmethod
    def get_marginal_from_probability_distribution(
            global_probability_distribution: np.ndarray,
            bits_of_interest: List[int],
            register_names: Optional[List[str]] = None) -> np.ndarray:

        """Return marginal distribution from vector of global distribution
        :param global_probability_distribution: distribution on all bits
        :param bits_of_interest: bits we are interested in (so we average over other bits)
                                Assuming that qubits are labeled
                                from 0 to log2(len(global_probability_distribution))
        :param register_names: bitstrings register, default is
                               '00...00', '000...01', '000...10', ..., etc.

        :return: marginal_distribution : marginal probability distribution

        NOTE: we identify bits with qubits in the variables bitstring_names
        """

        #JT:  Marginal function move to qrem.common.probability 
        return fda.get_marginal_from_probability_distribution(global_probability_distribution=global_probability_distribution,
                                                              bits_of_interest=bits_of_interest,
                                                              register_names=register_names)

    #JT: In this method an error occures when subset dictionary is restricted to some set of settings

    def compute_average_marginal_for_subset(self,
                                            subset):

        marginals_dictionary = self._marginals_dictionary

        marginal_dict_now = {}
        subset = tuple(subset)

        for input_state_bitstring, dictionary_marginals_now in marginals_dictionary.items():
            #JT: a marginal is constructed out of qubits belonging to subset
            input_marginal = ''.join([input_state_bitstring[x] for x in subset])
            #JT: computations of the marginal is perfored
            if subset not in dictionary_marginals_now.keys():
                #self.compute_unnormalized_marginals([input_state_bitstring], [subset])
                #my addition
                dictionary_marginals_now = self.compute_unnormalized_marginals([input_state_bitstring], [subset])[input_state_bitstring]
            if input_marginal not in marginal_dict_now.keys():
                marginal_dict_now[input_marginal] = dictionary_marginals_now[subset]
            else:
                marginal_dict_now[input_marginal] += dictionary_marginals_now[subset]

        for key_small in marginal_dict_now.keys():
            marginal_dict_now[key_small] /= np.sum(marginal_dict_now[key_small])

        return marginal_dict_now

    def get_averaged_marginal_for_subset(self,
                                         subset: Tuple[int]):

        #TODO FBM: I think it does not work here, different format
        #JT: Well said, it does not work indeed :)

        #marginals_dictionary = self._marginals_dictionary

        #if subset in marginals_dictionary.keys():
        #    return marginals_dictionary[subset]
        #else:
        #    return self.compute_average_marginal_for_subset(subset)
        
        #JT my proposition to change this
        marginals_dictionary = self._marginals_dictionary

        if subset in marginals_dictionary.values():
            return marginals_dictionary[subset]
        else:
            return self.compute_average_marginal_for_subset(subset)


        # subset_key = 'q' + 'q'.join([str(s) for s in subset])




#MOcomm This looks like outdated file - do we need it?
#ORGANIZE - probably this file should be delated

import pickle

from tqdm import tqdm
import numpy as np

from qrem.functions_qrem import ancillary_functions as anf
from qrem.functions_qrem import functions_data_analysis as fda
from qrem.noise_characterization.base_classes.marginals_analyzer_base import MarginalsAnalyzerBase
from qrem.noise_simulation.CN import sampling_improvements as si

from qrem.common.printer import qprint

# specify data used for testing
backend_name = 'ASPEN-9'
date = '2021-09-27'
number_of_qubits = 20
bitstrings_right_to_left = False

# Specify whether save calculated data
saving = True

running_estimation = False

module_directory = anf.get_local_storage_directory()
tests_directory_main = module_directory + '/saved_data/characterization2021/'
tests_directory_low = f"ground_states_implementation/2SAT/{date}/N{number_of_qubits}/"
directory_to_open = tests_directory_main + 'processed_data/' + tests_directory_low

with open(directory_to_open + "00_raw_results.pkl", 'rb') as filein:
    experiments_results = pickle.load(filein)

results_dictionary = experiments_results['results_dictionary']

with open(directory_to_open + "full_information_2SAT_N20.pkl", 'rb') as filein:
    hamiltonians_data = pickle.load(filein)

# results_dictionary = experiments_results['results_dictionary']


tests_directory_low = f"{date}/{backend_name}/N{number_of_qubits}/"
directory_to_open = tests_directory_main + 'processed_data/' + tests_directory_low
with open(directory_to_open + 'correction_data_tests.pkl', 'rb') as filein:
    dictionary_data_noise_models = pickle.load(filein)

joint_correction_matrices_dictionary = dictionary_data_noise_models[
    'joint_correction_matrices_dictionary']
noise_models_correction_indices = dictionary_data_noise_models['correction_indices']
joint_noise_matrices_dictionary = dictionary_data_noise_models['joint_noise_matrices_dictionary']

# print(noise_models_correction_indices[0])
total_number = 0
new_noise_model_correction_indices = []
# already_done_models_heh = []
all_models = {tup[0]: tup for tup in noise_models_correction_indices}
merged_noise_models = {}
already_done_models_heh = []
mapping_for_oskar = {}
for model_parameters, stuff in all_models.items():
    if model_parameters in already_done_models_heh:
        continue

    list_now = [model_parameters]
    already_done_models_heh.append(model_parameters)
    for model_parameters2, stuff2 in all_models.items():
        if model_parameters2 in already_done_models_heh:
            continue

        if model_parameters != model_parameters2:
            if stuff2[1] == stuff[1]:
                list_now.append(model_parameters2)
                already_done_models_heh.append(model_parameters2)

    merged_noise_models[tuple(list_now)] = (stuff[1], stuff[2])

noise_models_correction_indices = [(key, value[0], value[1]) for key, value in
                                   merged_noise_models.items()]

# anf.save_results_pickle(dictionary_to_save=noise_models_correction_indices,
#                         directory=directory_to_open,
#                         custom_name='sorted_clusters_data')


results_big_dictionary = {model_parameters: {} for model_parameters, _, _ in
                          noise_models_correction_indices}

all_pairs = [(i, j) for i in range(number_of_qubits) for j in
             range(i + 1, number_of_qubits)]

all_pairs = all_pairs + [(i,) for i in range(number_of_qubits)]

if running_estimation:

    # estimated_results = {}
    estimated_energies = {}
    estimated_marginals = {}
    for input_state in tqdm(list(results_dictionary.keys())):
        hamiltonian_dict = hamiltonians_data[input_state]['']
        weights_dictionary = hamiltonian_dict['weights_dictionary']
        needed_marginals = list(weights_dictionary.keys())
        # energy_ideal = hamiltonian_dict['ground_state_energy']

        results_now = results_dictionary[input_state]
        marginals_calculator = MarginalsAnalyzerBase(
            results_dictionary={input_state: results_now},
            bitstrings_right_to_left=bitstrings_right_to_left
        )

        marginals_calculator.compute_unnormalized_marginals(experiment_keys=[input_state],
                                                            subsets_dictionary=all_pairs)

        marginals_dictionary = marginals_calculator.marginals_dictionary[input_state]
        energy_estimated = fda.estimate_energy_from_marginals(
            weights_dictionary=weights_dictionary,
            marginals_dictionary=marginals_dictionary)

        estimated_energies[input_state] = energy_estimated
        estimated_marginals[input_state] = marginals_dictionary

    anf.save_results_pickle(dictionary_to_save=estimated_energies,
                            directory=directory_to_open,
                            custom_name='estimation_energy_results_all')

    anf.save_results_pickle(dictionary_to_save=estimated_marginals,
                            directory=directory_to_open,
                            custom_name='estimation_marginals_results_all')

else:
    with open(directory_to_open + 'estimation_energy_results_all.pkl', 'rb') as filein:
        estimated_energies = pickle.load(filein)

    with open(directory_to_open + 'estimation_marginals_results_all.pkl', 'rb') as filein:
        estimated_marginals = pickle.load(filein)

# raise KeyboardInterrupt
do_sanity_checks = False

test_cl = [(x,) for x in range(number_of_qubits)]

# print(register)
#
all_pairs = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]


# all_pairs = all_pairs + [(i,) for i in range(number_of_qubits)]


def bitstring_swap_wise(bistring, substring, bits):
    # t000=time.time()
    count_bad = [c for c in bistring]
    for b in range(len(substring)):
        count_bad[bits[b]] = substring[b]

    count_bad = count_bad

    returning_this = ''.join([c for c in count_bad])
    # qprint('swap wise took:',time.time()-t000)

    return returning_this


def create_marginal_bitstring(bitstring, bits_of_interest):
    # TODO: make sure about it
    # print(bitstring)
    reversed_order = bitstring
    return ''.join([reversed_order[b] for b in bits_of_interest])


def stochastic_noise_to_count(count,
                              bits,
                              mask,
                              error_matrix):
    random_mask = np.random.choice(mask, p=
    error_matrix[:, int(
        create_marginal_bitstring(count, bits), 2)])

    return bitstring_swap_wise(count, random_mask, bits)


def multiple_stochastic_noise_to_count(
        count,
        set_of_bits,
        masks,
        error_matrices):
    # assuming lack of correlations between error matrices
    new_string = count
    for i in range(len(set_of_bits)):
        new_string = stochastic_noise_to_count(new_string,
                                               set_of_bits[i],
                                               masks[i],
                                               error_matrices[i])

    return new_string


def get_multiple_marginal_bitstrings(count, set_of_bits):
    return [create_marginal_bitstring(count, bs) for bs in set_of_bits]


def create_masks(bits, size):
    return anf.register_names_qubits(bits, size, False)


def make_memory_noisy(memory, bits_to_flip, error_matrices):
    # N=len(memory)
    masks = [create_masks(range(len(bs)), len(bs)) for bs in bits_to_flip]

    return [multiple_stochastic_noise_to_count(count,
                                               bits_to_flip,
                                               masks,
                                               error_matrices)
            for count in memory]


from qrem.noise_mitigation.expected_values.CRS import SCR_functions as scr_fun

run_symmetrization_finders = False

if run_symmetrization_finders:
    symmetrizers_all = {}
    for model_parameters, noise_model_dictionary, _ in tqdm(noise_models_correction_indices[0:]):

        clusters_list = noise_model_dictionary['clusters_list']
        local_noise_matrices = {cluster: joint_noise_matrices_dictionary[cluster]['averaged'] for cluster
                                in clusters_list}

        qprint("\nNoise model now:",clusters_list)

        error_probabilitities_dictionary = {}
        symmetrizers_dictionary = {}

        for cluster in clusters_list:
            local_map = local_noise_matrices[cluster]
            if len(cluster)==1:
                equalizer = scr_fun.find_general_symmetrizer(stochastic_matrix=local_map
                                                             )
                equalized_matrix = equalizer@local_map
                error_probability = equalized_matrix[0,1]
                error_probabilitities_dictionary[cluster] = error_probability
            else:
                equalizer, errors_probabilities = scr_fun.find_optimal_decoupler(stochastic_matrix=local_map,
                                                                                 printing=False)
                for index_qubit in range(len(errors_probabilities)):
                    error_probabilitities_dictionary[(cluster[index_qubit],)] = errors_probabilities[index_qubit]

            symmetrizers_dictionary[cluster] = equalizer
        # numpy.set_printoptions(threshold=sys.maxsize)

        qprint("\nOptimal error probabilities:", error_probabilitities_dictionary)

        symmetrizers_all[model_parameters] = {'symmetrizers_dictionary':symmetrizers_dictionary,
                                              'error_probabilities':error_probabilitities_dictionary}

        # qprint("DONE!\n")
        print("\n")
        # qprint("error probabilities:")

    anf.save_results_pickle(dictionary_to_save=symmetrizers_all,
                            directory=directory_to_open,
                            custom_name='symmetrizers_dictionaries')

#
#
else:
    with open(directory_to_open + 'symmetrizers_dictionaries.pkl', 'rb') as filein:
        symmetrizers_all = pickle.load(filein)
#
from typing import List, Tuple

def sort_clusters_division(clusters_division:List[Tuple[int]]):

    sorted_inside = [tuple(sorted(x)) for x in clusters_division]

    return tuple(sorted(sorted_inside, key = lambda x: x[0]))



unique_correction_indices = {}
unique_cluster_divisions = list(np.unique([sort_clusters_division(noise_model_dictionary['clusters_list'])
                                           for _, noise_model_dictionary, __ in noise_models_correction_indices]))
# for model_parameters, noise_model_dictionary, _ in noise_models_correction_indices:

model_parameters_dict_unique_clusters = {cluster_division:[] for cluster_division in unique_cluster_divisions}

models_dict = {model_parameters:noise_model_dictionary for model_parameters, noise_model_dictionary, _ in noise_models_correction_indices}


for model_parameters, noise_model_dictionary, _ in noise_models_correction_indices:
    clusters_list = sort_clusters_division(noise_model_dictionary['clusters_list'])
    model_parameters_dict_unique_clusters[clusters_list].append(model_parameters)

original_model_parameters = {}
model_parameters_dict_unique_clusters_formatted_correctly = {}
for cluster_division, model_parameters in model_parameters_dict_unique_clusters.items():
    one_list = tuple([p for params in model_parameters for p in params])
    model_parameters_dict_unique_clusters_formatted_correctly[cluster_division] = (one_list, models_dict[model_parameters[0]])

    original_model_parameters[one_list] = model_parameters[0]

number_of_samples = 10 ** 5
test_08, test_18, test_78, test_01 = 0, 0, 0, 0

print(len(unique_cluster_divisions))


# noise_model_identifiers = {}
# counter = 0
# for clusters_list, (model_parameters, noise_model_dictionary) \
#         in list(model_parameters_dict_unique_clusters_formatted_correctly.items())[0:]:
#
#     noise_model_identifiers[model_parameters] = counter
#
#     counter+=1

#
# anf.save_results_pickle(dictionary_to_save=noise_model_identifiers,
#                         directory=directory_to_open,
#                         custom_name='noise_model_identifiers_updated')

with open(directory_to_open + 'noise_model_identifiers_updated.pkl', 'rb') as filein:
    noise_model_identifiers = pickle.load(filein)

run_index = 0
qprint("\nRUN INDEX:",run_index,'red')
print()
# for clusters_list, model_parameters, noise_model_dictionary in model_parameters_dict_unique_clusters_formatted_correctly.items()
for clusters_list, (model_parameters, noise_model_dictionary) \
        in list(model_parameters_dict_unique_clusters_formatted_correctly.items())[run_index*3:(run_index+1)*3]:


    noise_model_identifier_now = noise_model_identifiers[model_parameters]


    # clusters_list = noise_model_dictionary['clusters_list']
    qprint("\nCurrent noise model", clusters_list)
    print(model_parameters)
    # raise KeyboardInterrupt
    # print(clusters_list==_sort_clusters_division(noise_model_dictionary['clusters_list']))
    # print([len(x) for x in clusters_list])
    # clusters_list = [(2*j,2*j+1) for j in range(10)]
    # clusters_list = [(j,) for j in range(number_of_qubits)]
    # raise KeyboardInterrupt
    qubits_to_clusters_map = {}
    for qubit in range(number_of_qubits):
        for cluster in clusters_list:
            if qubit in cluster:
                qubits_to_clusters_map[qubit] = cluster


    local_noise_matrices = {cluster: joint_noise_matrices_dictionary[cluster]['averaged'] for cluster
                            in clusters_list}

    params_key = original_model_parameters[model_parameters]

    symmetrizers_dictionary = symmetrizers_all[params_key]['symmetrizers_dictionary']
    error_probabilitities_dictionary = symmetrizers_all[params_key]['error_probabilities']
    # raise KeyboardInterrupt

    range_keys = list(results_dictionary.keys())

    # TODO added starting from different place

    results_dictionary_models = {}
    figures_of_merit_list = []
    for input_state in tqdm(range_keys[0:]):
        energy_estimated = estimated_energies[input_state]
        hamiltonian_dict = hamiltonians_data[input_state]['']
        weights_dictionary = hamiltonian_dict['weights_dictionary']
        needed_marginals = list(weights_dictionary.keys())
        energy_ideal = hamiltonian_dict['ground_state_energy']

        results_now_dict = results_dictionary[input_state]


        unique_counts = list(results_now_dict.keys())
        probability_distro = np.array(list(results_now_dict.values()),dtype=float)
        numer_of_shots_in_experiments = sum(probability_distro)
        probability_distro/=numer_of_shots_in_experiments

        resampled_outcomes = np.random.multinomial(pvals=probability_distro,n=number_of_samples)


        resampled_outcomes_dictionary = {unique_counts[outcome_index]:resampled_outcomes[outcome_index]
                                         for outcome_index in range(len(resampled_outcomes))}

        equalized_samples = si.sample_from_multiple_input_states(input_states_dictionary=resampled_outcomes_dictionary,
                                                                 local_stochastic_matrices=symmetrizers_dictionary,
                                                                 print_progres_bar=False)


        energy_corrected = 0
        for subset in needed_marginals:
            local_energy = fda.estimate_energy_from_counts_dictionary(counts_dictionary=equalized_samples,
                                                                      weights_dictionary={subset:weights_dictionary[subset]})
            multiplier = 1

            for qi in subset:
                multiplier/=(1 - 2 * error_probabilitities_dictionary[(qi,)])

            energy_corrected+=multiplier*local_energy

        error_corrected = abs(energy_corrected-energy_ideal)/number_of_qubits
        error_estimated = abs(energy_estimated-energy_ideal)/number_of_qubits
        figure_of_merit = error_corrected/error_estimated
        figures_of_merit_list.append(figure_of_merit)

        # # qprint("Energy ideal:",energy_ideal)
        # qprint("\nEnergy estimated error per qubit:",np.round(error_estimated,3))
        # # qprint("Energy corrected:", energy_corrected)
        # qprint("Energy corrected error per qubit:", np.round(error_corrected,3))
        # qprint("Figure of merit", np.round(figure_of_merit,3))
        # print()

        # print(energy_ideal,
        #       energy_estimated,
        #       energy_corrected,
        #       abs(energy_corrected - energy_ideal) / number_of_qubits
        #       # en_noisy
        #       )

        # raise KeyboardInterrupt


        results_dictionary_models[input_state] = {
            'error_corrected': error_corrected,
            'error_estimated': error_estimated,
            'energy_corrected': energy_corrected,
            'energy_estimated': energy_estimated,
            'energy_ideal': energy_ideal,

        }

    qprint("Mean figure of merit:", np.mean(figures_of_merit_list))
    # raise KeyboardInterrupt

    if saving:
        anf.save_results_pickle(dictionary_to_save=results_dictionary_models,
                                directory=directory_to_open + '/SCR_models_mitigation/s_10^6/',
                                custom_name="id_" + str(noise_model_identifier_now))





"""Run AWS Experiment pipeline

This module allows to run a characterization/mitigation/benchmark experiment 
on a chosen AWS braket quantum machine. It uses a config file, which example can be found in 
<configs/ default_aws.ini> file. Read the config to understand avialable parameters.


 documentation as specified by the `NumPy
Documentation HOWTO`_. Docstrings may extend over multiple lines. Sections
are created with a section header followed by an underline of equal length.

Example
-------
You can run experiment by (i) providing a given config via -C argument, for example

    $ python run_aws_experiment.py -C \\path\\to\\experiment\\config.ini

    
Notes
-----
    @authors: 
    @contact:
"""


import os, sys
import qrem
import pickle

from pathlib import Path

import numpy as np

from qrem.common.io import date_time_formatted
from qrem.common.printer import qprint, errprint, warprint
from qrem.common.experiment import tomography as tomography
from qrem.types import CircuitCollection
from qrem.common.constants import CIRCUIT_DATA_TYPE as c_type
from qrem.common.providers import aws_braket
# ----------------------------------------------------------------
# this are the parameters of running qrem in future / maybe move to config[GENERAL]?:
# you will be able to run it asa command
# ----------------------------------------------------------------
CONFIG_PATH = ["--config_file", "/media/tuzjan/T7/work_tuzjan/experiment_settings/aws_rigetti_coherence_experiment/aws_rigetti_coherence.ini"]

#TODO (PP) write version of runner that works with existing circuit collection file

def run( cmd_args=CONFIG_PATH, verbose_log = True,idle_run = False): 
    # ----------------------------------------------------------------
    #[1] We load an already defined default config
    # ----------------------------------------------------------------
    config = qrem.load_config(cmd_args = cmd_args, verbose_log = verbose_log)

    EXPERIMENT_NAME = config.experiment_name

    EXPERIMENT_FOLDER_PATH = Path(config.experiment_path)
    if not EXPERIMENT_FOLDER_PATH.is_dir():
        EXPERIMENT_FOLDER_PATH.mkdir( parents=True, exist_ok=True )

    BACKUP_CIRCUITS = config.backup_circuits
    BACKUP_CIRCUITS_METADATA = config.backup_circuits_metadata
    JOB_TAGS = list(config.job_tags)



    # ----------------------------------------------------------------
    #[2] Get info from provider about valid qubits
    # ----------------------------------------------------------------

    aws_device, metadata = aws_braket.get_device(device_full_name = config.device_name, verbose_log = config.verbose_log);
    valid_qubit_properties = aws_braket.get_valid_qubits_properties(device=aws_device, threshold=None, verbose_log = config.verbose_log)#config.gate_threshold, verbose_log = config.verbose_log)

    METADATA = metadata
    METADATA["date"] = date_time_formatted()
    METADATA["JOB_TAGS"] = JOB_TAGS
    if BACKUP_CIRCUITS_METADATA:
        METADATA["valid_qubit_properties"] = valid_qubit_properties

    number_of_qubits = valid_qubit_properties["number_of_good_qubits"]
    good_qubits_indices = valid_qubit_properties["good_qubits_indices"]

    if config.ground_states_circuits_path and config.number_of_ground_states != None:
        circuits_ground_states_preparation_collection = []

        GROUND_STATES_PATH = config.ground_states_circuits_path
        GROUND_STATES_NUMBER = config.number_of_ground_states
        with open(GROUND_STATES_PATH, 'rb') as filein:
            hamiltonians_data_dictionary= pickle.load(filein)
            GROUND_STATES_NUMBER = min(config.number_of_ground_states,len(hamiltonians_data_dictionary))

            for i in range(GROUND_STATES_NUMBER):
                ground_state = hamiltonians_data_dictionary[i]['ground_state']
                if i==0:
                    circuits_ground_states_preparation_collection=np.array([ground_state])
                else:
                    circuits_ground_states_preparation_collection=np.append(circuits_ground_states_preparation_collection,np.array([ground_state]),axis=0)

    # ----------------------------------------------------------------
    #[3] TODO (PP) check if there are specific qubits subset defined in config,. if yes - check against good_qubits_indices and choose only the subset
    # ----------------------------------------------------------------

    # ----------------------------------------------------------------
    #[4] Generate circuits
    # ----------------------------------------------------------------
    #[4.0] Setup circuit collection object
    qrem_circuit_collection = CircuitCollection(EXPERIMENT_NAME)
    qrem_circuit_collection.load_config(config=config)
    qrem_circuit_collection.qubit_indices = good_qubits_indices
    qrem_circuit_collection.metadata = METADATA

    #[4.1] TODO (PP) finish implementation of generate_circuits, so you know how many shots per circuit you need
    # - TODO (PP) initialize with config object
    qrem_circuit_collection.circuits, _, theoretical_total_circuit_count, theoretical_number_of_shots = tomography.generate_circuits(   number_of_qubits = number_of_qubits,
                                                                            experiment_type = config.experiment_type, 
                                                                            k_locality = config.k_locality,
                                                                            limited_circuit_randomness = config.limited_circuit_randomness,
                                                                            imposed_max_random_circuit_count = config.random_circuits_count,
                                                                            imposed_max_number_of_shots = config.shots_per_circuit)
    


    # ----------------------------------------------------------------
    #[4.1] Add coherence witness circuits
    # ----------------------------------------------------------------
    # 

    

    if config.coherence_witness_circuits:

        COHERENCE_WITNESS_CIRCUITS_PATH = Path(config.coherence_witness_circuits_path)

        with open(COHERENCE_WITNESS_CIRCUITS_PATH, 'rb') as filein:
            coherence_witness_circuits =  pickle.load(filein)

        qrem_circuit_collection.circuits = np.append(qrem_circuit_collection.circuits,coherence_witness_circuits,axis=0)

        qrem_circuit_collection.experiment_type = 'QDOT'




    # ----------------------------------------------------------------
    #[4.2] Add ground state circuits
    # ----------------------------------------------------------------
    # - TODO (PP) does not check correctness of circuits now and of pickle object
    
    if config.ground_states_circuits_path and config.number_of_ground_states != 0:

                
        qrem_circuit_collection.circuits = np.append(qrem_circuit_collection.circuits,circuits_ground_states_preparation_collection,axis=0)
    
    # ----------------------------------------------------------------
    #[4.3] Order of circuits is rearranged to ensure that benchmark and 
    # coherence witness circuits are run in the same availability window ""
    # ----------------------------------------------------------------


    #number_of_coherence_and_benchmark_circuits = len(coherence_witness_circuits) + GROUND_STATES_NUMBER 

    #temporary_rearrangement_array = copy.copy(qrem_circuit_collection.circuits[-number_of_coherence_and_benchmark_circuits:])

    #qrem_circuit_collection.circuits[-number_of_coherence_and_benchmark_circuits:] = qrem_circuit_collection.circuits[100:100+number_of_coherence_and_benchmark_circuits]

    #qrem_circuit_collection.circuits[100:100+number_of_coherence_and_benchmark_circuits] = temporary_rearrangement_array



    # - TODO (PP) save final number of shots to collection
    # qrem_circuit_collection.no_shots = ?

    #[5] At this stage, circuits are ready, we can save them to file
    if BACKUP_CIRCUITS:    
        qrem_circuit_collection.export_json(str(EXPERIMENT_FOLDER_PATH.joinpath("input_circuit_collection.json")), overwrite = False)

    #[6] Now we need to use aws_braket module to prepare circuits in aws_braket format
    braket_circuits = aws_braket.translate_circuits_to_braket_format(qrem_circuit_collection,valid_qubit_indices=good_qubits_indices)

    SUBMISSION_FOLDER_PATH = Path(config.experiment_path).joinpath("job_submission")

    if not SUBMISSION_FOLDER_PATH.is_dir():
        EXPERIMENT_FOLDER_PATH.mkdir( parents=True, exist_ok=True )
    
    total_number_of_circuits = len(qrem_circuit_collection.circuits) 
    print(f"Total number of circuits: {total_number_of_circuits} ")


    #[6] Now we need to prepare and run circuits
    circuits_ready = aws_braket.prepare_cricuits( braket_circuits = braket_circuits,
                     circuit_collection = qrem_circuit_collection,
                     good_qubits_indices = good_qubits_indices,
                     number_of_repetitions = config.shots_per_circuit,
                     number_of_task_retries = config.aws_braket_task_retries,
                     experiment_name = EXPERIMENT_NAME,
                     job_tags = JOB_TAGS,
                     pickle_submission = config.aws_pickle_results,
                     metadata = METADATA,
                     verbose_log = config.verbose_log,
                     job_dir = SUBMISSION_FOLDER_PATH,
                     overwrite_output = False)
    


    #TODO UNCOMMENT TO RUN
    if not circuits_ready:
        print("ERROR during circuit creation, aborting.")
    else:
        if not idle_run:
            aws_braket.execute_circuits( device_name=config.device_name,
                            pickle_submission = config.aws_pickle_results,
                            job_dir=SUBMISSION_FOLDER_PATH)



if __name__ == "__main__":
    TEST_RUN=False
    warprint("=============================")
    warprint("START: Executing AWS-BRAKET Experiment pipeline")
    warprint("=============================")
    argv = sys.argv[1:]
    import time
    t0 = time.time()

    if argv != []:
        run(cmd_args=argv, idle_run = TEST_RUN)
    else:
        run(idle_run = TEST_RUN)

    t1 = time.time()
    qprint("=============================")
    qprint("Elapsed time: {} sec".format(t1 - t0))


# ----------------------------------------------------------------
# how to run:
# ----------------------------------------------------------------
    #python run_aws_experiment.py -c path/to/config.ini
    #python run_aws_experiment.py
















#------------------------------------
#------------------------------------
#------------------------------------
#------------------------------------
#qrem.run_characterisation_experiment()
#? qrem.run_mitigation_experiment()
#qrem.download_characterisation_experiment_results()
#qrem.download_mitigation_experiment_results()
#qrem.run_characterisation()
#qrem.run_mitigation()




# -*- coding: utf-8 -*-
"""
Created on Wed Nov  10 11:24:48 2015

@author: Peter Wittek
"""
from __future__ import division, print_function
from math import sqrt
from sympy import S
from sympy.physics.quantum.dagger import Dagger
from .sdp_relaxation import SdpRelaxation
from .nc_utils import apply_substitutions, is_number_type, ncdegree, \
    separate_scalar_factor


class RdmHierarchy(SdpRelaxation):

    """Class for obtaining a level in the reduced density matrix method
    :param variables: Commutative or noncommutative, Hermitian or nonhermiatian
                      variables, possibly a list of list of variables if the
                      hierarchy is not NPA.
    :type variables: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`
                     or a list of list.
    :param parameters: Optional symbolic variables for which moments are not
                       generated.
    :type parameters: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`
                     or a list of list.
    :param verbose: Optional parameter for level of verbosity:

                       * 0: quiet
                       * 1: verbose
                       * 2: debug level
    :type verbose: int.
    :param normalized: Optional parameter for changing the normalization of
                       states over which the optimization happens. Turn it off
                       if further processing is done on the SDP matrix before
                       solving it.
    :type normalized: bool.
    :param ppt: Optional parameter to impose a partial positivity constraint
                on the moment matrix.
    :type ppt: bool.

    Attributes:
      - `monomial_sets`: The monomial sets that generate the moment matrix blocks.

      - `monomial_index`: Dictionary that maps monomials to SDP variables.

      - `constraints`: The complete set of constraints after preprocesssing.

      - `primal`: The primal optimal value.

      - `dual`: The dual optimal value.

      - `x_mat`: The primal solution matrix.

      - `y_mat`: The dual solution matrix.

      - `solution_time`: The amount of time taken to solve the relaxation.

      - `status`: The solution status of the relaxation.
    """

    def __init__(self, variables, parameters=None, verbose=0, circulant=False,
                 parallel=False):
        super(RdmHierarchy, self).__init__(variables, parameters, verbose,
                                           False, parallel)
        self.circulant = circulant
        self.m_block = 0

    def _push_monomials(self, monomials, n_vars, row_offset, coords, N):
        monomial0 = apply_substitutions(monomials[0], self.substitutions,
                                        self.pure_substitution_rules)
        for mon1 in monomials[1:]:
            self.moment_substitutions[mon1] = monomial0
        k, coeff = [], []
        if is_number_type(monomial0):
            coeff = [monomial0]
            k = [0]
        elif monomial0.is_Add:
            for element in monomial0.as_ordered_terms():
                n_vars, k1, coeff1 = self._push_monomials([element], n_vars,
                                                          row_offset, coords,
                                                          N)
                k += k1
                coeff += coeff1

        elif monomial0 != 0:
            entries = self._process_monomial(monomial0, n_vars)
            for entry in entries:
                k1, coeff1 = entry
                k.append(k1)
                coeff.append(coeff1)
        for rowA, columnA in coords:
            for ki, coeffi in zip(k, coeff):
                self.F[row_offset + rowA * N + columnA, ki] = coeffi
        if max(k) > n_vars:
            n_vars = max(k)
        return n_vars, k, coeff

    def __second_moments(self, n_vars, monomialsA, block_index,
                         processed_entries):
        N = len(monomialsA)
        row_offset = 0
        if block_index > 0:
            for block_size in self.block_struct[0:block_index]:
                row_offset += block_size ** 2
        coords, mons =  \
            generate_block_coords(monomialsA[:N // 2], monomialsA[:N // 2],
                                  0, N // 2, 0, N // 2, 0, 0, N // 2)
        coords_, mons_ = \
            generate_block_coords(monomialsA[:N // 2], monomialsA[N // 2:],
                                  N // 2, N, 0, N // 2, 0, 0, N)
        coords += coords_
        mons += mons_
        coords_, mons_ = \
            generate_block_coords(monomialsA[N // 2:], monomialsA[N // 2:],
                                  N // 2, N, N // 2, N, 0, 0, N)
        coords += coords_
        mons += mons_
        for mon, coord in zip(mons, coords):
            n_vars, _, _ = self._push_monomials(mon, n_vars, row_offset,
                                                coord, len(monomialsA))
        return n_vars, block_index + 1, processed_entries

    def __fourth_moments(self, n_vars, monomialsA, block_index,
                         processed_entries):
        self.m_block += 1
        if self.m_block == 1 or self.m_block == 3:
            N = int(sqrt(len(monomialsA)))
            row_offset = 0
            if block_index > 0:
                for block_size in self.block_struct[0:block_index]:
                    row_offset += block_size ** 2
            for block_row in range(N):
                monsA = monomialsA[N*block_row:N*(block_row+1)]
                for block_col in range(block_row, N):
                    monsB = monomialsA[N*block_col:N*(block_col+1)]
                    coords, mons = \
                        generate_block_coords(monsA, monsB, 0, N, 0, N,
                                              N*block_row, N*block_col, N)
                    for mon, coord in zip(mons, coords):
                        n_vars, _, _ = self._push_monomials(mon, n_vars,
                                                            row_offset,
                                                            coord, N**2)
            return n_vars, block_index + 1, processed_entries
        elif self.m_block == 2:
            N = int(sqrt(len(monomialsA)//2))
            row_offset = 0
            if block_index > 0:
                for block_size in self.block_struct[0:block_index]:
                    row_offset += block_size ** 2
            for block_row in range(N):
                monsA = monomialsA[N*block_row:N*(block_row+1)]
                for block_col in range(block_row, N):
                    monsB = monomialsA[N*block_col:N*(block_col+1)]
                    coords, mons = \
                        generate_block_coords(monsA, monsB, 0, N, 0, N,
                                              N*block_row, N*block_col, N)
                    for mon, coord in zip(mons, coords):
                        n_vars, _, _ = self._push_monomials(mon, n_vars,
                                                            row_offset,
                                                            coord, 2*N**2)

            for block_row in range(N):
                monsA = monomialsA[N*block_row:N*(block_row+1)]
                for block_col in range(N):
                    monsB = monomialsA[N**2+N*block_col:N**2+N*(block_col+1)]
                    coords, mons = \
                        generate_block_coords(monsA, monsB, 0, N, 0, N,
                                              N*block_row,
                                              N**2 + N*block_col, N)
                    for mon, coord in zip(mons, coords):
                        n_vars, _, _ = self._push_monomials(mon, n_vars,
                                                            row_offset,
                                                            coord, 2*N**2)

            for block_row in range(N):
                monsA = monomialsA[N**2+N*block_row:N**2+N*(block_row+1)]
                for block_col in range(block_row, N):
                    monsB = monomialsA[N**2+N*block_col:N**2+N*(block_col+1)]
                    coords, mons = \
                        generate_block_coords(monsA, monsB, 0, N, 0, N,
                                              N**2 + N*block_row,
                                              N**2 + N*block_col, N)
                    for mon, coord in zip(mons, coords):
                        n_vars, _, _ = self._push_monomials(mon, n_vars,
                                                            row_offset,
                                                            coord, 2*N**2)
            return n_vars, block_index + 1, processed_entries
        else:
            return super(RdmHierarchy, self).\
                    _generate_moment_matrix(n_vars, block_index,
                                            processed_entries,
                                            monomialsA, [S.One])

    def _generate_moment_matrix(self, n_vars, block_index, processed_entries,
                                monomialsA, monomialsB, ppt=False):
        if self.circulant:
            if ncdegree(monomialsA[0]) == 1:
                return self.__second_moments(n_vars, monomialsA, block_index,
                                             processed_entries)
            elif ncdegree(monomialsA[0]) == 2:
                return self.__fourth_moments(n_vars, monomialsA, block_index,
                                             processed_entries)
            else:
                raise Exception("Cannot generate circulant moment matrix with "
                                "degree-" + str(ncdegree(monomialsA[0])) +
                                " terms.")

        else:
            return super(RdmHierarchy, self).\
                    _generate_moment_matrix(n_vars, block_index,
                                            processed_entries,
                                            monomialsA, monomialsB)


def generate_block_coords(monomialsA, monomialsB, col0, colN, row0, rowN,
                          row_offset, col_offset, row_length):
    coords, mons = [], []
    for column in range(col0, colN):
        coords.append([(row_offset + row, col_offset + column + row - row0)
                       for row in range(row0, rowN)
                       if column + row - row0 < row_length])
        mons.append([Dagger(monomialsA[row-row_offset-row0]) *
                     monomialsB[col-col_offset-col0]
                     for row, col in coords[-1]])
        if row_offset != col_offset or row0 < col0:
            lower_triangular = [(col - col0 - col_offset + row_offset + row0,
                                 row - row0 - row_offset + col_offset + col0)
                                for row, col in coords[-1]
                                if row - row0 - row_offset !=
                                col - col0 - col_offset]
            if lower_triangular != []:
                lower_mons = [Dagger(monomialsA[row-row_offset-row0]) *
                              monomialsB[col-col_offset-col0]
                              for row, col in lower_triangular]
                coords.append(lower_triangular)
                mons.append(lower_mons)
    return coords, mons



import numpy as np
import copy
from typing import Optional, List, Dict, Union, Tuple

from tqdm import tqdm

from qrem.functions_qrem import ancillary_functions as anf

#JT: Check whwter we need this here
from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.functions.functions_noise_model_heuristic import partition_algorithm_v1_cummulative

from qrem.noise_characterization.data_analysis.InitialNoiseAnalyzer import InitialNoiseAnalyzer


from qrem.common.printer import qprint
from qrem.common import convert, utils

class NoiseModelGenerator(InitialNoiseAnalyzer):
    """
        This is class that uses results of Diagonal Detector Tomography (DDT) to construct stochastic_matrix noise
        model for a measuring device. The model is classical and based on Ref. [0.5].

        The main functionalities include computing sets of strongly correlated qubits (clusters)
        and for each such set, computing the qubits which affect the exact form of the noise
        on those clusters (hence, neighborhoods of cluster)
    """

    def __init__(self,
                 results_dictionary: Dict[str, Dict[str, int]],
                 bitstrings_right_to_left: Optional[bool] = False,
                 marginals_dictionary: Optional[Dict[str, Dict[Tuple[int], np.ndarray]]] = None,
                 noise_matrices_dictionary: Optional[
                     Dict[Tuple[int], Dict[Union[str,Tuple[int]], Dict[str, np.ndarray]]]] = None,
                 clusters_list: Optional[List[Tuple[int]]] = None,
                 neighborhoods: Dict[Tuple[int], Tuple[int]] = None,
                 correlations_data:Optional[np.ndarray] = None
                 ) -> None:

        super().__init__(results_dictionary=results_dictionary,
                         bitstrings_right_to_left=bitstrings_right_to_left,
                         marginals_dictionary=marginals_dictionary,
                         noise_matrices_dictionary=noise_matrices_dictionary,
                         correlations_data=correlations_data
                         )

        # self._correlations_table_pairs = correlations_table_pairs

        if clusters_list is None:
            clusters_list = []

        if neighborhoods is None:
            neighborhoods = {}

        self._clusters_list = clusters_list

        self._neighborhoods = neighborhoods

    @property
    def clusters_list(self) -> List[Tuple[int]]:
        return self._clusters_list

    @clusters_list.setter
    def clusters_list(self, clusters_list: List[Tuple[int]]) -> None:
        for cluster in clusters_list:
            # cluster_string = anf.get_qubits_keystring(cluster)
            cluster_string = tuple(cluster)
            if cluster_string not in self._noise_matrices_dictionary.keys():
                average_noise_matrix_now = self._compute_noise_matrix_averaged(cluster)
                dictionary_now = {'averaged': average_noise_matrix_now}

                if cluster_string in self._neighborhoods.keys():
                    neighborhood_now = self._neighborhoods[cluster_string]
                    dependent_noise_matrices = self._compute_noise_matrix_dependent(cluster,
                                                                                    neighborhood_now)
                    dictionary_now = {**dictionary_now, **dependent_noise_matrices}
                    qprint('im doing this')

                self._noise_matrices_dictionary[tuple(cluster)] = dictionary_now

        self._clusters_list = clusters_list

    @property
    def neighborhoods(self) -> Dict[Tuple[int], Tuple[int]]:
        return self._neighborhoods

    @neighborhoods.setter
    def neighborhoods(self, neighborhoods: Dict[Tuple[int], Tuple[int]]) -> None:
        self._neighborhoods = neighborhoods
        self.clusters_list = [cluster for cluster in neighborhoods.keys()]

        for cluster_string in neighborhoods.keys():
            dictionary_now = self._noise_matrices_dictionary[cluster_string]
            neighborhood_now = neighborhoods[cluster_string]
            # print(dictionary_now.keys())
            neighbors_key = anf.get_qubits_keystring(neighborhood_now)
            if neighbors_key not in dictionary_now.keys():
                cluster = convert.get_qubit_indices_from_keystring(cluster_string)

                dependent_noise_matrices = self._compute_noise_matrix_dependent(cluster,
                                                                                neighborhood_now)

                self._noise_matrices_dictionary[cluster_string] = {**dictionary_now,
                                                                   **dependent_noise_matrices}

    def _compute_clusters_pairwise(self,
                                   maximal_size: int,
                                   cluster_threshold: float
                                   ) -> list:
        """
            Get partition of qubits in a device into disjoint "clusters". This function uses "naive"
            method_name by assigning qubits to the same cluster if correlations between them are higher
            than some "neighbors_threshold". It restricts size of the cluster to "maximal_size"
            by disregarding the lowest correlations (that are above neighbors_threshold).
            It uses table of correlations from class property self._correlations_table_pairs

          :param cluster_threshold: correlations magnitude above which qubits are assigned
                 to the same cluster
          :param maximal_size: maximal allowed size of the cluster

          :return: clusters_labels_list: list of lists, each representing a single cluster
             """
        self._clusters_list = []

        qubit_indices = self._qubit_indices
        # number_of_qubits = len(qubit_indices)

        clusters = {'q%s' % qi: [[qi, 0., 0.]] for qi in qubit_indices}
        for qi in qubit_indices:
            for qj in qubit_indices:
                if qj > qi:
                    corr_j_i, corr_i_j = self._correlations_table[qj, qi], \
                                         self._correlations_table[qi, qj]

                    # if any of the qubit affects the other strongly enough,
                    # we assign them to the same cluster
                    if corr_j_i >= cluster_threshold or corr_i_j >= cluster_threshold:
                        clusters['q%s' % qi].append([qj, corr_i_j, corr_j_i])
                        clusters['q%s' % qj].append([qi, corr_i_j, corr_j_i])

        # Merge clusters containing the same qubits
        #(PP) this seems quite inefficient approach

        #TODO BEGIN improve efficiency of this check
        new_lists = []

        for key, value in clusters.items():
            clusters[key] = sorted(value, key=lambda arg: arg[0])
            new_lists.append([vi[0] for vi in clusters[key]])

        # print(new_lists)
        while utils.check_for_multiple_occurences(new_lists):
            for i in range(len(new_lists)):
                cl0 = new_lists[i]
                for j in range(len(new_lists)):
                    cl1 = new_lists[j]
                    if len(utils.lists_intersection(cl0, cl1)) != 0:
                        new_lists[i] = utils.lists_sum(cl0, cl1)

            uniques = np.unique(new_lists)

            if not isinstance(uniques[0], list):
                uniques = [uniques]

            unique_stuff = [sorted(lis) for lis in uniques]
            new_lists = copy.deepcopy(unique_stuff)
        #END ------------------------------------------------

        clusters_list = new_lists


        # Chop clusters if they exceed max size
        while np.any(np.array([len(cluster) > maximal_size for cluster in clusters_list])):
            chopped_clusters = []
            for cluster in clusters_list:
                if len(cluster) > maximal_size:
                    correlations_sorting = []
                    for qi in cluster:
                        # as figure of merit, we will sum all correlations that are between
                        # given qubit and other guys in its cluster.
                        x = 0.0
                        for list_now in clusters['q%s' % qi]:
                            x += np.max([list_now[1], list_now[2]])

                        correlations_sorting.append([qi, x])

                    correlations_sorted = sorted(correlations_sorting,
                                                 key=lambda arg: arg[1],
                                                 reverse=True)

                    # choose only "maximal_size" qubits to belong to given cluster
                    qubits_sorted = [correlations_sorted[index][0] for index in range(maximal_size)]
                    qubits_left = [correlations_sorted[index][0] for index in
                                   range(maximal_size, len(correlations_sorted))]
                    chopped_clusters.append(sorted(qubits_left))
                else:
                    qubits_sorted = cluster
                chopped_clusters.append(sorted(qubits_sorted))

            clusters_list = sorted(chopped_clusters, key=lambda y: y[0])

        self._clusters_list = clusters_list

        return clusters_list

    def _find_neighbors_of_cluster_holistic(self,
                                            cluster: List[int],
                                            maximal_size: int,
                                            chopping_threshold: Optional[float] = 0.) -> List[int]:
        """
        For a given cluster of qubits, find qubits which are their neighbors, i.e., they affect the
        noise matrix of cluster significantly. Figure of merit for correlations here is:

        c_{j -> cluster} = 1/2 || Lambda_{cluster}^{Y_j='0'}- Lambda_{cluster}^{Y_j='1'}||_{l1}

        where Lambda_{cluster}^{Y_j} is the noise matrix describing noise on qubits in "cluster"
        provided that input state of qubit "j" was "Y_j".
        See also description of self._compute_clusters_pairwise.


        :param cluster: list of labels of qubits in a cluster
        :param maximal_size: maximal allowed size of the set "cluster+neighborhood"
        :param chopping_threshold: numerical value, for which correlations lower than
              chopping_threshold are set to 0.
              If not provided, it adds all_neighbors until maximal_size is met.



        :return: neighbors_list: list of lists, each representing a single cluster
        """

        size_cut = maximal_size - len(cluster)

        potential_neighbours = []

        #JT this is added as a treshold

        correlation_threshold = 0.009

        for qi in self._qubit_indices:
            if qi not in cluster:
                compute_noise_matrix = False

                #JT this is added to avoid extensive computation of noise matrices

                for qj in cluster: 
                    corr_i_j = self._correlations_data['worst_case']['classical'][qj, qi]

                    #
                    if corr_i_j > correlation_threshold:
                        
                        compute_noise_matrix = True
                        
                        break

                #if the corresponding correlation coefficient is above threshold computation of the noise matrix will be performed 
                if compute_noise_matrix:    
                    lam_ci_j = self.get_noise_matrix_dependent(cluster,
                                                           [qi])
                    diff_ci_j = lam_ci_j['0'] - lam_ci_j['1']
                    correlation_ci_j = 1 / 2 * np.linalg.norm(diff_ci_j, ord=1)
                    potential_neighbours.append([qi, correlation_ci_j])

        sorted_neighbours = sorted(potential_neighbours, key=lambda x: x[1], reverse=True)

        neighbors_list = sorted(
            [sorted_neighbours[i][0] for i in range(int(np.min([size_cut, len(sorted_neighbours)]))) if
             chopping_threshold < sorted_neighbours[i][1]])

        cluster_key = anf.get_qubits_keystring(cluster)

        self._neighborhoods[cluster_key] = neighbors_list

        return neighbors_list

    def _find_all_neighborhoods_holistic(self,
                                         maximal_size,
                                         chopping_threshold: float,
                                         show_progress_bar: Optional[bool] = False) \
            -> Dict[str, List[int]]:
        """
                Run self._find_neighbors_of_cluster_holistic for all clusters.

                :param maximal_size: maximal allowed size of the set "cluster+neighborhood"
                :param chopping_threshold: numerical value, for which correlations lower than
                      chopping_threshold are set to 0.
                      If not provided, it adds all_neighbors until maximal_size is met.
                :param show_progress_bar: specify whether to show progress bar


                :return: neighbors_dictionary: dictionary where KEY is label for cluster,
                                               and VALUE is list of its neighbors
        """

        self._neighborhoods = {}
        clusters_list = self._clusters_list
        range_clusters = range(len(clusters_list))

        if show_progress_bar:
            range_clusters = tqdm(range_clusters)

        for index_cluster in range_clusters:
            cluster = clusters_list[index_cluster]
            self._neighborhoods[
                anf.get_qubits_keystring(cluster)] = self._find_neighbors_of_cluster_holistic(
                cluster,
                maximal_size,
                chopping_threshold)

        return self._neighborhoods

    def _find_neighbors_of_cluster_pairwise(self,
                                            cluster: List[int],
                                            maximal_size: int,
                                            neighbors_threshold: float
                                            ) -> List[int]:
        """
        Like self._find_neighbors_of_cluster_holistic but looks how noise on qubits in given cluster
        depend on input state of other qubits (potential neighbors) *separately*.

        NOTE: see description of self._find_neighbors_of_cluster_holistic
                                            for definition of correlations' measure we use

        :param cluster: list of labels of qubits in a cluster
        :param maximal_size: maximal allowed size of the set "cluster+neighborhood"
        :param neighbors_threshold: numerical value, for which correlations higher than
              neighbors_threshold assign qubit to the neighborhood of other qubit

        :return: neighbors_list: list of lists, each representing a single cluster
        """

        qubit_indices = self._qubit_indices
        potential_neighbors = []

        for qj in qubit_indices:
            affections_qj = []
            for qi in cluster:
                if qj not in cluster:
                    corr_j_i = self._correlations_table[qi, qj]
                    affections_qj.append(corr_j_i)

            if qj not in cluster:
                corr_j_i = np.max(affections_qj)

                if corr_j_i >= neighbors_threshold:
                    potential_neighbors.append([qj, corr_j_i])
        sorted_neighbors = sorted(potential_neighbors, key=lambda x: x[1], reverse=True)

        target_size = maximal_size - len(cluster)
        range_final = int(np.min([len(sorted_neighbors), target_size]))

        return sorted([sorted_neighbors[index][0] for index in
                       range(range_final)])

    def _find_all_neighborhoods_pairwise(self,
                                         maximal_size: int,
                                         neighbors_threshold: float,
                                         show_progress_bar: Optional[bool] = False
                                         ) -> Dict[str, List[int]]:
        """
        Like self._find_neighbors_of_cluster_holistic but looks how noise on qubits in given cluster
        depend on input state of other qubits (potential neighbors) *separately*.

        NOTE: see description of self._find_neighbors_of_cluster_holistic
                                            for definition of correlations' measure we use

        :param maximal_size: maximal allowed size of the set "cluster+neighborhood"
        :param neighbors_threshold: numerical value, for which correlations higher than
              neighbors_threshold assign qubit to the neighborhood of other qubit

        :return: neighbors_dictionary: dictionary where KEY is label for cluster, and VALUE is list of
                                       its neighbors
        """
        if self._correlations_table is None:
            self._compute_correlations_data_pairs_old()
#why do we use old version of the method if we have a new one?
        self._neighborhoods = {}

        clusters_list = self._clusters_list
        range_clusters = range(len(clusters_list))

        if show_progress_bar:
            range_clusters = tqdm(range_clusters)

        for index_cluster in range_clusters:
            cluster = clusters_list[index_cluster]
            self._neighborhoods[tuple(cluster)] = self._find_neighbors_of_cluster_pairwise(
                cluster, maximal_size=maximal_size, neighbors_threshold=neighbors_threshold)

        return self._neighborhoods

    def compute_clusters(self,
                         maximal_size: int,
                         correlations_table:np.ndarray,
                         method: Optional[str] = 'CSP',
                         method_kwargs: Optional[dict] = None) -> list:
        """
        Get partition of qubits in a device into disjoint "clusters".
        This function uses various heuristic methods, specified via string "version".
        It uses table of correlations from class property self._correlations_table_pairs

        :param maximal_size: maximal allowed size of the cluster
        :param method: string specifying stochasticity_type of heuristic
        Possible values:
            'pairwise' - heuristic that uses Algorithm 3 from Ref.[]
            'CSP' - heuristic Cluster Size Penalization


        :param method_kwargs: potential arguments that will be passed to clustering function.
                           For possible parameters see descriptions of particular functions_qrem.
        :return: clusters_labels_list: list of lists, each representing a single cluster
        """

        self._clusters_list = []

        if method == 'pairwise':
            if method_kwargs is None:
                default_kwargs = {'maximal_size': maximal_size,
                                  'cluster_threshold': 0.02
                                  }

                method_kwargs = default_kwargs
            elif 'maximal_size' in method_kwargs.keys():
                if method_kwargs['maximal_size'] != maximal_size:
                    raise ValueError('Disagreement between maximal size argument and method_name kwargs')
            else:
                method_kwargs['maximal_size'] = maximal_size
            clusters_list = self._compute_clusters_pairwise(**method_kwargs)


        elif method.upper()=='CSP':

            from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.child_classes.heuristic_clustering_algorithm import HeuristicClusteringAlgorithm

            if method_kwargs is None:
                method_kwargs = {'maximal_size': maximal_size,
                                  'version': 'v4',
                                  'number_of_algorithm_iterations':1000,
                                  'alpha_multiplier':1.0
                                  }

                # method_kwargs = default_kwargs

            if 'version' not in method_kwargs.keys():
                method_kwargs['version'] = 'v4'

            if 'number_of_algorithm_iterations' not in method_kwargs.keys():
                method_kwargs['number_of_algorithm_iterations'] = 1000

            if 'alpha_multiplier' not in method_kwargs.keys():
                method_kwargs['alpha_multiplier'] = 1.0

            if 'maximal_size' not in method_kwargs.keys():
                method_kwargs['maximal_size'] = maximal_size

            #v0 is interpreted sa "pairwise"
            if method_kwargs['version'] == 'v0':

                clustering_algorithm = HeuristicClusteringAlgorithm(
                    correlations_table=correlations_table,
                    version='v1',
                    C_maxsize=maximal_size)

                clustering_algorithm.suggest_alpha_heuristic()

                alpha = 2*clustering_algorithm.alpha*method_kwargs['alpha_multiplier']
                kwargs_pairwise = {'maximal_size': maximal_size,
                                  'cluster_threshold': alpha
                                  }

                clusters_list = self._compute_clusters_pairwise(**kwargs_pairwise)

                self._alpha_parameter = alpha

            #JT Here clustering algorithm is chosen

            else:
                clustering_algorithm = HeuristicClusteringAlgorithm(correlations_table=correlations_table,
                                                                    version=method_kwargs['version'],
                                                                    C_maxsize=maximal_size,
                                                                    N_alg = method_kwargs['number_of_algorithm_iterations'])#, N_rand = method_kwargs['N_rand'])

                clustering_algorithm.suggest_alpha_heuristic()

                # print(clustering_algorithm.alpha)
                #TODO someone: make sure that this 2 factor is consitent with the paper
                clustering_algorithm.alpha = 2*clustering_algorithm.alpha*method_kwargs['alpha_multiplier']


                clusters_list, score = clustering_algorithm.clusterize()

                # qprint('Current partitioning got score:', score)

                self._alpha_parameter = clustering_algorithm.alpha




        elif method == 'holistic_v1':
            if method_kwargs is None:
                alpha = 1
                algorithm_runs = 1000
                default_kwargs = {'alpha': alpha,
                                  'N_alg': algorithm_runs,
                                  'printing': False,
                                  'drawing': False,
                                  'C_maxsize': maximal_size}

                method_kwargs = default_kwargs

            elif 'C_maxsize' in method_kwargs.keys():
                # TODO FBM, OS: this variable should have name consistent with rest of functions_qrem
                if method_kwargs['C_maxsize'] != maximal_size:
                    raise ValueError('Disagreement between maximal size argument and method_name kwargs')
            else:
                method_kwargs['C_maxsize'] = maximal_size

            clusters_list, score = partition_algorithm_v1_cummulative(correlations_table,
                                                                      **method_kwargs)

            qprint('Current partitioning got score:', score)
        elif method == 'external':
            if 'clustering_algorithm' not in method_kwargs.keys():
                raise ValueError('No HeuristicClusteringAlgorithmBase instance provided in method kwargs')
            clustering_algorithm = method_kwargs['clustering_algorithm']
            if clustering_algorithm.C_maxsize != maximal_size:
                raise ValueError(
                    'Disagreement between maximal size argument and method_name kwargs')
            clusters_list, score = clustering_algorithm.clusterize()
        else:
            raise ValueError('No heuristic with that name: ' + method)

        clusters_list_formatted = [tuple(sorted(x)) for x in clusters_list]
        clusters_list_formatted = tuple(sorted(clusters_list_formatted, key=lambda x: x[0]))
        from qrem.functions_qrem import functions_data_analysis as fda

        clusters_list_formatted = fda._sort_clusters_division(clusters_list_formatted)

        self._clusters_list = clusters_list_formatted

#apart from clusters list also value of the clustering function is returned
        return clusters_list_formatted, score

    def find_all_neighborhoods(self,
                               maximal_size: int,
                               method: Optional[str] = 'holistic',
                               method_kwargs: Optional[dict] = None):

        if method == 'pairwise':
            if method_kwargs is None:
                default_kwargs = {'neighbors_threshold': 0.01}
                method_kwargs = default_kwargs

            method_kwargs['maximal_size'] = maximal_size
            neighborhoods = self._find_all_neighborhoods_pairwise(**method_kwargs)

        elif method == 'holistic':
            if method_kwargs is None:
                default_kwargs = {'chopping_threshold': 0.01,
                                  'show_progress_bar': True}
                method_kwargs = default_kwargs
            method_kwargs['maximal_size'] = maximal_size
            neighborhoods = self._find_all_neighborhoods_holistic(**method_kwargs)

        else:
            raise ValueError('Wrong method_name name')

        return neighborhoods
#JT : This can be safely removed 
    def print_properties(self):
        # TODO FBM, OS: add this

        return None

    def draw_noise_model(self):
        # TODO FBM, OS: add this

        return None



"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""

from qrem.functions_qrem.functions_noise_model_heuristic import partition_algorithm_v1_cummulative
from qrem.functions_qrem.functions_noise_model_heuristic import partition_algorithm_v2_cummulative
from qrem.functions_qrem.functions_noise_model_heuristic import partition_algorithm_v3_cummulative
from qrem.functions_qrem.functions_noise_model_heuristic import partition_algorithm_v4_cummulative


class HeuristicClusteringAlgorithmBase:

    def __init__(self, version=None, correlations_table=None, alpha=None, C_maxsize=None, N_alg=None, N_rand=None,
                 printing=False, drawing=False, disable_pb=False):
        self.version = version
        self.correlations_table = correlations_table
        self.alpha = alpha
        self.C_maxsize = C_maxsize
        self.N_alg = N_alg
        self.N_rand = N_rand
        self.printing = printing
        self.drawing = drawing
        self.disable_pb = disable_pb
        self.global_best_parition_sorted = None
        self.global_best_value = None

    def clusterize(self, correlations_table=None, alpha=None):
        if correlations_table is not None:
            self.correlations_table = correlations_table
        elif self.correlations_table is None:
            raise ValueError("Correlations_table not specified.")
        if alpha is not None:
            self.alpha = alpha
        elif self.alpha is None:
            raise ValueError("Alpha not specified.")
        if self.C_maxsize is None:
            raise ValueError("C_maxsize not specified.")
        if self.N_alg is None:
            raise ValueError("N_alg not specified.")

        if self.version == 'v1':
            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v1_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.printing,
                self.drawing, self.disable_pb)
        elif self.version == 'v2':
            if self.N_rand is None:
                raise ValueError("N_rand not specified.")

            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v2_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.N_rand,
                self.printing,
                self.drawing, self.disable_pb)

        elif self.version == 'v3':
            if self.N_rand is None:
                raise ValueError("N_rand not specified.")

            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v3_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.N_rand,
                self.printing,
                self.drawing, self.disable_pb)
        elif self.version == 'v4':

            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v4_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.printing,
                self.drawing, self.disable_pb)
        else:
            raise ValueError("Implemented versions are: v1, v2, v3, v4")

        return self.global_best_parition_sorted, self.global_best_value



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

from typing import Dict, Tuple, Optional

import numpy as np
from qrem.functions_qrem import ancillary_functions as anf, functions_data_analysis as fda
from qrem.ctmp.modeltools.ncpol2sdpa import generate_variables, flatten, SdpRelaxation
from sympy import S


def generate_commuting_measurements(party,
                                    label):
    measurements = []
    for i in range(len(party)):
        measurements.append(generate_variables(label + '%s' % i,
                                               party[i] - 1,
                                               hermitian=True))
    return measurements


def find_ground_state_energy_bruteforce(weights_dictionary,
                                        number_of_qubits):
    classical_states = anf.get_classical_register_bitstrings(qubit_indices=range(number_of_qubits),
                                                             )

    ground_state_energy = 10 ** 6
    ground_state = None
    for bitstring in classical_states:

        energy_now = fda.get_energy_from_bitstring_diagonal_local(bitstring=bitstring,
                                                                  weights_dictionary=weights_dictionary)

        if energy_now < ground_state_energy:
            ground_state_energy = energy_now
            ground_state = bitstring

    return ground_state_energy, ground_state


def get_symbolic_hamiltonian_from_weights_dictionary(weights_dictionary: Dict[Tuple[int], float],
                                                     spins):

    spins_list = flatten(spins)
    hamiltonian = 0
    # print(spins_list)
    for qubits_subset, weight in weights_dictionary.items():
        if weight != 0:
            if len(qubits_subset) == 1:
                qi = qubits_subset[0]
                hamiltonian += float(weight) * spins_list[qi]
            elif len(qubits_subset) == 2:
                (qi, qj) = qubits_subset
                hamiltonian += float(weight) * spins_list[qi] * spins_list[qj]


            # if len(qubits_subset) == 1:
            #     qi = qubits_subset[0]
            #     hamiltonian += weight * spins[qi][0][0][0]
            # elif len(qubits_subset) == 2:
            #     (qi, qj) = qubits_subset
            #     hamiltonian += weight * spins[qi][0][0][0] * spins[qj][0][0][0]

    return hamiltonian


def __extract_ground_state_from_moments_matrix(moments_matrix,
                                               number_of_qubits):
    dimension = moments_matrix.shape[0]
    # number_of_qubits = int((dimension-1)/2)
    # print(number_of_qubits)

    spin_configuration = []

    for index_now in range(number_of_qubits):

        sign = np.sign(moments_matrix[index_now + 1, 0])
        if sign == -1:
            spin_now = 1
        elif sign == 1:
            spin_now = 0
        else:
            raise ValueError()
        spin_configuration.append(spin_now)

    return spin_configuration


from qrem.functions_qrem.CBB import branchandbound_tools as bb_tools, spin_models as spin_tools
from qrem.ctmp.modeltools.ncpol2sdpa.chordal_extension import find_variable_cliques


def find_ground_state_with_CBB(weights_dictionary: Dict[Tuple[int], float],
                               number_of_qubits: int,
                               verbose:Optional[int]=0):

    # configuration = [2 for _ in range(1)]
    # spin_variables = [generate_commuting_measurements(configuration, chr(65 + i))
    #          for i in range(number_of_qubits)]
    # print(spin_variables)

    # length = number_of_qubits
    # number_of_qubits = 2
    configuration = [2]
    spin_variables = spin_tools.get_square2D_spins(k=number_of_qubits,
                                                   l=1,
                                                   configuration=configuration)
    substitutions = {M ** 2: S.One for M in flatten(spin_variables)}

    # spin_variables = spin_tools.get_square2D_spins(k=number_of_qubits,
    #                                                l=1,
    #                                                configuration=configuration)
    #
    # local = spin_tools.get_2Dsquare_localdisorder(k=number_of_qubits, l=1, sigma=0.5)
    # hamiltonian_polynomial = spin_tools.get_2Dsquare_ferromagneticdisorder_hamiltonian(k=number_of_qubits, l=1,
    #                                                                         local=local,
    #                                                                         s_variables=spin_variables)
    #


    # print(hamiltonian_polynomial)
    #
    # #
    hamiltonian_polynomial = get_symbolic_hamiltonian_from_weights_dictionary(weights_dictionary=weights_dictionary,
                                                                   spins=spin_variables)

    # hamiltonian_polynomial = spin_variables[0][0][0][0]*1
    # #
    # print(hamiltonian_polynomial)


    cliques = find_variable_cliques(variables=flatten(spin_variables),
                                    objective=hamiltonian_polynomial)

    # print(cliques)

    z_low, z_up, ground_state_candidate_list = bb_tools.get_groundBandB(s_variables=spin_variables,
                                                                   substitutions=substitutions,
                                                                   hamiltonian=hamiltonian_polynomial,
                                                                   cliques=cliques,
                                                                   verbose=verbose)


    ground_state_candidate = ''.join(['1' if s==-1 else '0' for s in ground_state_candidate_list]
                                     )




    return z_low, z_up, ground_state_candidate


def find_ground_state_with_SDP_relaxations(weights_dictionary: Dict[Tuple[int], float],
                                           number_of_qubits: int,
                                           hierarchy_level: Optional[float] = 2,
                                           get_also_upper_bound=True,

                                           return_ground_state_approximation=False):
    # number_of_qubits = weights_dictionary.shape[0]
    # generates the symbolic spin variables
    configuration = [2 for _ in range(1)]
    spins = [generate_commuting_measurements(configuration, chr(65 + i))
             for i in range(number_of_qubits)]

    # gets the hamiltonian in the form of a polynomial in the spins
    hamiltonian_polynomial = get_symbolic_hamiltonian_from_weights_dictionary(weights_dictionary, spins)

    # properties of the spin operators: square equal to identity
    substitutions = {M ** 2: S.One for M in flatten(spins)}

    # print(spins)
    # generating the sdp moment matrix
    sdp = SdpRelaxation(flatten(spins), verbose=2)
    sdp.get_relaxation(hierarchy_level,
                       substitutions=substitutions,
                       objective=hamiltonian_polynomial)

    # solving for the lower bound
    sdp.solve(solver="mosek")
    low = sdp.primal

    if sdp.status is not 'optimal':
        print('ERROR!!! The status is not optimal')

    # print(low)
    # print(sdp.__dict__.keys())
    # print(sdp.block_struct)

    if return_ground_state_approximation:
        ground_state_candidate = __extract_ground_state_from_moments_matrix(
            moments_matrix=sdp.x_mat[0],
            number_of_qubits=number_of_qubits)
        ground_state_candidate = ''.join([str(s) for s in ground_state_candidate])

    # solving for the upper bound
    if get_also_upper_bound:
        sdp.set_objective(objective=-hamiltonian_polynomial)

        sdp.solve(solver="mosek")
        up = -sdp.primal
    else:
        up = None


    if sdp.status is not 'optimal':
        print('ERROR!!! The status is not optimal')

    if return_ground_state_approximation:
        return low, up, ground_state_candidate
    else:
        return low, up

#
#
# raise KeyboardInterrupt


# ||\Lambda^{-1}||_{1-->1}

# 5q, 5q, 5q, 5q, 5q
# lam0 \otimes lam1 \otimes lam2

# q0, q6, q11
# exp(15) ---> exp(3)

# 2q, 2q
# lam0 \otimes lam1
# 2q 2q -------------------------> 1q, 1q, 1q, 1q
# A0 * lam0 \otimes A1 lam1 = s0 \otimes s1 \otimes s2 \otimes s3



import numpy as np
from typing import Optional
from collections import defaultdict
from sympy import S
from qrem.ctmp.modeltools.ncpol2sdpa.sdp_relaxation import SdpRelaxation
from qrem.ctmp.modeltools.ncpol2sdpa.nc_utils import generate_variables, flatten
from qrem.functions_qrem.functions_benchmarks import create_hamiltonians_for_benchmarks


from typing import Dict, Tuple, List

"""Glue function to access old code
"""
def estimate_gamma(n: int, rates: List[Tuple]) -> np.float:
    # convert rates to format from FM's code
    rates_dict = {}
    for error in rates:
        i, j, error_in, _, rate = error
        if i==j:
            key = (i,)
        else:
            key = (i, j)
        if key not in rates_dict:
                rates_dict[key] = defaultdict(float)
        rates_dict[key][error_in] = rate
    _, gamma = _find_optimal_noise_strength_CTMP_SDP_relaxation(rates_dict, n)
    return gamma

"""Generates a collectinon of ground state approximations for random Hamiltonians together with ground state 
energy lower and upper bound.

Returns a list of tuples (state, low, up), where:
state - string describing the approximate ground state
low, up - lower and upper bound on the ground state energy
h - Hamiltonian
"""
def generate_random_ground_states(n_qubits: int, n_hamiltonians: int, clause_density: float) -> List[Tuple]:
    result = []
    random_hamiltonians = create_hamiltonians_for_benchmarks(n_qubits, n_hamiltonians, clause_density)
    for h in random_hamiltonians.values():
        weights = h['weights_dictionary']
        low, _, state = find_ground_state_with_SDP_relaxations(weights, n_qubits, return_ground_state_approximation=True)
        mat = h['weights_matrix'].copy()
        d = np.diag(mat).copy()
        np.fill_diagonal(mat, 0)
        vec = np.array([-2*float(v)+1 for v in state])
        # once state Psi is found, upper bound on the ground state energy by <Psi|H|Psi>
        up = np.dot(vec.T, np.dot(mat, vec)) + np.dot(d.T, vec)
        result.append((state, low, up, h))
    return result


"""Code below copy pasted from Filip's code
"""

def generate_commuting_measurements(party, label):
    measurements = []
    for i in range(len(party)):
        measurements.append(generate_variables(label + '%s' % i,
                                               party[i] - 1,
                                               hermitian=True))
    return measurements


def get_symbolic_hamiltonian_from_weights_dictionary(weights_dictionary: Dict[Tuple[int], float],
                                                     spins):

    spins_list = flatten(spins)
    hamiltonian = 0
    for qubits_subset, weight in weights_dictionary.items():
        if weight != 0:
            if len(qubits_subset) == 1:
                qi = qubits_subset[0]
                hamiltonian += float(weight) * spins_list[qi]
            elif len(qubits_subset) == 2:
                (qi, qj) = qubits_subset
                hamiltonian += float(weight) * spins_list[qi] * spins_list[qj]
    return hamiltonian

def find_ground_state_with_SDP_relaxations(weights_dictionary: Dict[Tuple[int], float],
                                           number_of_qubits: int,
                                           hierarchy_level: Optional[float] = 2,
                                           get_also_upper_bound=True,
                                           return_ground_state_approximation=False,
                                           solver_name="cvxpy"):
    configuration = [2 for _ in range(1)]
    spins = [generate_commuting_measurements(configuration, chr(65 + i))
             for i in range(number_of_qubits)]
    hamiltonian_polynomial = get_symbolic_hamiltonian_from_weights_dictionary(weights_dictionary, spins)
    substitutions = {M ** 2: S.One for M in flatten(spins)}
    sdp = SdpRelaxation(flatten(spins), verbose=2)
    sdp.get_relaxation(hierarchy_level,
                       substitutions=substitutions,
                       objective=hamiltonian_polynomial)
    sdp.solve(solver=solver_name)
    #sdp.solve(solver="mosek")
    low = sdp.primal
    if sdp.status is not 'optimal':
        print('ERROR!!! The status is not optimal')
    if return_ground_state_approximation:
        ground_state_candidate = __extract_ground_state_from_moments_matrix(
            moments_matrix=sdp.x_mat[0],
            number_of_qubits=number_of_qubits)
        ground_state_candidate = ''.join([str(s) for s in ground_state_candidate])
    if get_also_upper_bound:
        sdp.set_objective(objective=-hamiltonian_polynomial)

        #sdp.solve(solver="mosek")
        sdp.solve(solver=solver_name)
        up = -sdp.primal
    else:
        up = None
    if sdp.status is not 'optimal':
        print('ERROR!!! The status is not optimal')
    if return_ground_state_approximation:
        return low, up, ground_state_candidate
    else:
        return low, up

def __convert_rates_dictionary_into_hamiltonian(ctmp_rates_dictionary,
                                                number_of_qubits):
    constant = 0
    hamiltonian_weights = defaultdict(float)
    for qubits_subset, rates_dic in ctmp_rates_dictionary.items():
        if len(qubits_subset) == 1:
            r0, r1 = rates_dic['0'], rates_dic['1']
            constant += (r0 + r1) / 2
            hamiltonian_weights[qubits_subset] += (r0 - r1) / 2
        elif len(qubits_subset) == 2:
            r00, r01, r10, r11 = rates_dic['00'], rates_dic['01'], rates_dic['10'], rates_dic['11']
            constant += (r00 + r01 + r10 + r11) / 4
            hamiltonian_weights[(qubits_subset[0],)] += (r00 + r01 - r10 - r11) / 4
            hamiltonian_weights[(qubits_subset[1],)] += (r00 - r01 + r10 - r11) / 4
            hamiltonian_weights[qubits_subset] += (r00 - r01 - r10 + r11) / 4
    return hamiltonian_weights, constant

def _find_optimal_noise_strength_CTMP_SDP_relaxation(
        ctmp_rates_dictionary: Dict[Tuple[int], Dict[str, float]],
        number_of_qubits: int,
        hierarchy_level: Optional[int] = 1) -> Tuple[float]:
    hamiltonian_dictionary, constant = __convert_rates_dictionary_into_hamiltonian(
        ctmp_rates_dictionary=ctmp_rates_dictionary,
        number_of_qubits=number_of_qubits)
    lower_bound, upper_bound = find_ground_state_with_SDP_relaxations(
        weights_dictionary=hamiltonian_dictionary,
        number_of_qubits=number_of_qubits,
        hierarchy_level=hierarchy_level)
    return lower_bound + constant, upper_bound + constant

def __extract_ground_state_from_moments_matrix(moments_matrix,
                                               number_of_qubits):
    dimension = moments_matrix.shape[0]
    # number_of_qubits = int((dimension-1)/2)
    # print(number_of_qubits)

    spin_configuration = []

    for index_now in range(number_of_qubits):

        sign = np.sign(moments_matrix[index_now + 1, 0])
        if sign == -1:
            spin_now = 1
        elif sign == 1:
            spin_now = 0
        else:
            raise ValueError()
        spin_configuration.append(spin_now)

    return spin_configuration




from tqdm import tqdm

import numpy as np
import matplotlib.pyplot as plt
import statistics
from qrem.functions_qrem import ancillary_functions as anf
from qrem.functions_qrem import functions_hamiltonians
from qrem.functions_qrem import functions_data_analysis as fdt

from qrem.backends_support.qiskit import qiskit_utilities
from qrem.noise_characterization.base_classes.OverlappingTomographyBase import OverlappingTomographyBase
from qrem.noise_characterization.tomography_design.overlapping.DOTMarginalsAnalyzer import DOTMarginalsAnalyzer
from qrem.noise_characterization.tomography_design.overlapping.SeparableCircuitsCreator import \
    SeparableCircuitsCreator
#MOcomm - it seams we are using two MarginalAnalizer classes to compute the same thing
from qrem.noise_characterization.tomography_design.overlapping.QDTMarginalsAnalyzer import \
    QDTMarginalsAnalyzer

from qrem.noise_mitigation.probability_distributions.MarginalsCorrector import MarginalsCorrector
from qrem.noise_simulation.CN import functions_sampling as fus
from qrem.noise_simulation.CN.noise_implementation import  simulate_noise_results_dictionary


from qrem.common import probability, math
from qrem.common.printer import qprint

import qrem.functions_qrem.povmtools as pt
from qrem import cn





#this is mostly


#function generating 2sat hamilonians for benchmarks
def create_hamiltonians_for_benchmarks(number_of_qubits, number_of_hamiltonians, clause_density):
    hamiltonian_name = '2SAT'
    hamiltonian_ordering_indices = list(range(0, number_of_hamiltonians))
    index_first, index_last = hamiltonian_ordering_indices[0], hamiltonian_ordering_indices[-1]
    # whether to attempt to solve Hamiltonians. WARNING: for big problems might be infeasible
    solve_hamiltonians = False
    all_hamiltonians = {}

    for hamiltonian_index in hamiltonian_ordering_indices:

        hamiltonian_data_now = functions_hamiltonians.generate_random_2SAT_hamiltonian(
            number_of_qubits=number_of_qubits,
            clause_density=clause_density)

        if solve_hamiltonians:
            hamiltonian_data_now = functions_hamiltonians.solve_2SAT_hamiltonian(hamiltonian_data=hamiltonian_data_now)

        all_hamiltonians[hamiltonian_index] = hamiltonian_data_now
    return all_hamiltonians





def create_eigenstates_benchmarks(number_of_qubits, number_of_states, noise_model):
    OT_creator = OverlappingTomographyBase(number_of_qubits=number_of_qubits,
                                           experiment_name='DDOT',
                                           maximal_circuits_number=number_of_states)

    circuits_QDOT = OT_creator.get_random_circuits_list(number_of_circuits=number_of_states)

    experiment_name = 'DDOT'
    circuits_creator = SeparableCircuitsCreator(SDK_name='qiskit',
                                                qubit_indices=range(number_of_qubits),
                                                quantum_register_size=5,
                                                classical_register_size=5,
                                                descriptions_of_circuits=circuits_QDOT,
                                                experiment_name=experiment_name)

    OT_circuits_list = circuits_creator.get_circuits()
    batches = anf.create_batches(circuits_list=OT_circuits_list, circuits_per_job=300)
    jobs_list = qiskit_utilities.run_batches(backend_name='qasm_simulator',
                                             batches=batches,
                                             shots=10 ** 4)

    unprocessed_results = common.providers.ibmutils.data_converters.get_counts_from_qiskit_jobs(jobs_list=jobs_list)

    processed_results = fdt.convert_counts_overlapping_tomography(counts_dictionary=unprocessed_results,
                                                                  experiment_name=experiment_name,
                                                                  reverse_bitstrings=True)
    noisy_results_dictionary = simulate_noise_results_dictionary(processed_results, noise_model[0], noise_model[1])

    subsets_of_qubits = subsets_of_qubits = math.get_k_local_subsets(number_of_elements=5,subset_size=2, all_sizes_up_to_k = True)
#MOcomm - we use here a very "low level class" - can't we use basic classes like marginals_analyzer_base? Janek?
    marginals_analyzer = QDTMarginalsAnalyzer(results_dictionary=noisy_results_dictionary,
                                              experiment_name=experiment_name)

    marginals_analyzer.compute_all_marginals(subsets_dictionary=subsets_of_qubits, show_progress_bar=True,
                                             multiprocessing=False)

    return noisy_results_dictionary, marginals_analyzer.marginals_dictionary


def eigenstate_energy_calculation_and_estimation(results_dictionary,marginals_dictionary, hamiltonians_data):
    results_energy_estimation = {}

    for state_index, input_state in tqdm(enumerate(marginals_dictionary.keys())):
        # Read Hamiltonian data
        hamiltonian_data_dictionary = hamiltonians_data[state_index]
        weights_dictionary = hamiltonian_data_dictionary['weights_dictionary']

        # Calculate ideal energy
        energy_ideal = fdt.get_energy_from_bitstring_diagonal_local(bitstring=input_state,
                                                                    weights_dictionary=weights_dictionary)

        # Read experimental results data
        marginals_dictionary_raw = marginals_dictionary[input_state]
        results_dictionary_raw = results_dictionary[input_state]

        missing_subsets = [x for x in weights_dictionary.keys() if x not in marginals_dictionary_raw.keys()]
        if len(missing_subsets) > 0:
            marginals_analyzer = DOTMarginalsAnalyzer({input_state: results_dictionary_raw})
            marginals_analyzer.compute_all_marginals(missing_subsets,
                                                     show_progress_bar=False,
                                                     multiprocessing=False)

            marginals_dictionary_raw = {**marginals_dictionary_raw,
                                        **marginals_analyzer.marginals_dictionary[input_state]}

        # Calculate experimentally estimated energy
        energy_raw = fdt.estimate_energy_from_marginals(weights_dictionary=weights_dictionary,
                                                        marginals_dictionary=marginals_dictionary_raw)

        dictionary_results_now = {'energy_ideal': energy_ideal,
                                  'energy_raw': energy_raw,
                                  'input_state': input_state,
                                  'weights_dictionary': weights_dictionary,
                                  'hamiltonian_index': state_index,
                                  }

        results_energy_estimation[state_index] = dictionary_results_now

    return results_energy_estimation

#JT: argument updated 
def run_benchmarks(number_of_qubits, results_dictionary, marginals_dictionary, energy_estimation_results, hamiltonians_data,
                   all_clusters_sets_dictionary, correction_matrices_dictionary, correction_indices_dictionary,mitigation_data_dictionary,noise_matrices_dictionary):
    noisy_energy_prediction = False
    noise_mitigation = True

    mitigation_method = 'lambda_inverse'
    method_kwargs = {'ensure_physicality': False}

    results_benchmarks = {state_index: {'energies': {}, 'errors': {}}
                          for state_index, input_state in enumerate(marginals_dictionary.keys())}

    
    #JT: This is a temporary addition, a dictionary that allows to test mitigation in an easier way, to be deleated

    #JT: Added for test purposes
    mitigated_marginals_dictionary_debugging={}

    
    for state_index, input_state in tqdm(enumerate(marginals_dictionary.keys())):
        #     print(f"{state_index}")

        
        energy_ideal, energy_raw = energy_estimation_results[state_index]['energy_ideal'], \
                                   energy_estimation_results[state_index]['energy_raw']

        energies_dictionary_now = {'energy_ideal': energy_ideal,
                                   'energy_raw': energy_raw}

        error_raw_now = abs(energy_ideal - energy_raw) / number_of_qubits

        errors_dictionary_now = {'error_raw': error_raw_now}

        # Read Hamiltonian data
        hamiltonian_data_dictionary = hamiltonians_data[state_index]
        weights_dictionary = hamiltonian_data_dictionary['weights_dictionary']
        needed_pairs = [x for x in weights_dictionary.keys() if len(x) == 2]

        # Read experimental results data
        marginals_dictionary_raw = marginals_dictionary[input_state]
        results_dictionary_raw = results_dictionary[input_state]

        missing_subsets = [x for x in weights_dictionary.keys() if x not in marginals_dictionary_raw.keys()]
        if len(missing_subsets) > 0:
            marginals_analyzer = DOTMarginalsAnalyzer({input_state: results_dictionary_raw})
            marginals_analyzer.compute_all_marginals(missing_subsets,
                                                     show_progress_bar=False,
                                                     multiprocessing=False)

            marginals_dictionary_raw = {**marginals_dictionary_raw,
                                        **marginals_analyzer.marginals_dictionary[input_state]}

        if noisy_energy_prediction:
            energies_dictionary_now['predicted_energies'] = {}
            errors_dictionary_now['energy_prediction_errors'] = {}

            for clusters_assignment, __ in all_clusters_sets_dictionary.items():
                noise_matrices_dictionary_model = {cluster: noise_matrices_dictionary[cluster]['averaged']
                                                   for cluster in clusters_assignment}

                predicted_energy = fus.get_noisy_energy_product_noise_model(input_state=input_state,
                                                                            noise_matrices_dictionary=noise_matrices_dictionary_model,
                                                                            needed_pairs=needed_pairs,
                                                                            weights_dictionary_tuples=weights_dictionary)

                energies_dictionary_now['predicted_energies'][clusters_assignment] = predicted_energy

                error_prediction = abs(predicted_energy - energy_raw) / number_of_qubits
                errors_dictionary_now['energy_prediction_errors'][clusters_assignment] = error_prediction

        if noise_mitigation:
            energies_dictionary_now['corrected_energies'] = {}
            errors_dictionary_now['corrected_errors'] = {}
            
            #the line below is added 
            mitigated_marginals_dictionary_debugging_now={}

            for clusters_assignment in all_clusters_sets_dictionary.keys():

                #marginals_dictionary_raw = marginals_dictionary[input_state]
                correction_indices_now = correction_indices_dictionary[clusters_assignment]
                correction_indices_now = mitigation_data_dictionary[clusters_assignment][0]
                correction_matrices_dictionary_new = mitigation_data_dictionary[clusters_assignment][1] 

                noise_matrices_dictionary_model = {cluster: noise_matrices_dictionary[cluster]['averaged']
                                                   for cluster in clusters_assignment}
                
                #JT: This needs to be changed
                correction_data_now = {'correction_indices': correction_indices_now,
                                       'correction_matrices_old': correction_matrices_dictionary,
                                       'correction_matrices': correction_matrices_dictionary_new, 
                                       'noise_matrices': noise_matrices_dictionary_model}

                missing_subsets = []

                for needed_subset in weights_dictionary.keys():
                    subset_for_correction = correction_indices_now[needed_subset]
             
                    if subset_for_correction not in marginals_dictionary_raw.keys():
                        missing_subsets.append(subset_for_correction)

                if len(missing_subsets) > 0:
                    marginals_analyzer = DOTMarginalsAnalyzer({input_state: results_dictionary_raw})
                    marginals_analyzer.compute_all_marginals(missing_subsets,
                                                             show_progress_bar=False,
                                                             multiprocessing=False)

                    marginals_dictionary_raw = {**marginals_dictionary_raw,
                                                **marginals_analyzer.marginals_dictionary[input_state]}

                marginals_corrector = MarginalsCorrector(
                    experimental_results_dictionary={input_state: results_dictionary_raw},
                    correction_data_dictionary=correction_data_now,
                    marginals_dictionary=marginals_dictionary_raw)

                marginals_dictionary_to_correct = {correction_indices_now[subset_to_correct]:
                                                       marginals_dictionary_raw[
                                                           correction_indices_now[subset_to_correct]]
                                                   for subset_to_correct in weights_dictionary.keys()}

                marginals_corrector.correct_marginals(marginals_dictionary=marginals_dictionary_to_correct,
                                                      method_kwargs=method_kwargs,
                                                      method=mitigation_method
                                                      )

                # Coarse-grain some of the corrected marginals_dictionary to obtain the ones that appear in Hamiltonian
                marginals_coarse_grained_corrected = \
                    marginals_corrector.compute_marginals_of_marginals(
                        weights_dictionary.keys(),
                        corrected=True)

                for key, marg in marginals_coarse_grained_corrected.items():
                    if not probability.is_valid_probability_vector(marg):
                        marginals_coarse_grained_corrected[key] = pt.find_closest_prob_vector_l2(marg).flatten()


                #the line below is added for debugging purposes
                mitigated_marginals_dictionary_debugging_now[clusters_assignment] = marginals_coarse_grained_corrected 
                
                energy_corrected = fdt.estimate_energy_from_marginals(weights_dictionary=weights_dictionary,
                                                       marginals_dictionary=
                                                       marginals_coarse_grained_corrected)
                #params_tuple = tuple(params)
               # print(type(clusters_assignment_data))
               # clusters_assignment_data = tuple([clusters_assignment,params_tuple])
                clusters_assignment_data = clusters_assignment
                energies_dictionary_now['corrected_energies'][clusters_assignment_data ] = energy_corrected

                error_correction = abs(energy_corrected - energy_ideal) / number_of_qubits
                errors_dictionary_now['corrected_errors'][clusters_assignment_data]= error_correction

        results_benchmarks[state_index]['energies'] = energies_dictionary_now
        results_benchmarks[state_index]['errors'] = errors_dictionary_now

        #JT: This is added for test purposes
        mitigated_marginals_dictionary_debugging[input_state]=mitigated_marginals_dictionary_debugging_now

    error_dictionary ={}
    for state_index in results_benchmarks.keys():
        for cluster_assigment,error in  results_benchmarks[state_index]['errors']['corrected_errors'].items():
            new_entry = {state_index:error}
            if cluster_assigment in error_dictionary.keys():
                error_dictionary[cluster_assigment] = {**error_dictionary[cluster_assigment],**new_entry}
            else:
                error_dictionary[cluster_assigment] =  new_entry
    
    for cluster_assigment, state_errors in error_dictionary.items():
        error_dictionary[cluster_assigment] = {**state_errors, **{'median': statistics.median(list(state_errors.values())), 'mean': statistics.mean(list(state_errors.values()))}}
        

     

      



    qprint("ALL DONE!", '', 'red')
    return results_benchmarks, error_dictionary#,   mitigated_marginals_dictionary_debugging


def calculate_mitigated_energy_for_hamiltonian(benchmark_hamiltonian, mitigated_marginals_dictionary):
    weights_dictionary = benchmark_hamiltonian['weights_dictionary']
    mitigated_energy = fdt.estimate_energy_from_marginals(weights_dictionary=weights_dictionary,
                                                       marginals_dictionary=
                                                       mitigated_marginals_dictionary)
    return mitigated_energy

def run_mitigation_benchmark_for_hamiltonian(benchmark_hamiltonian, results_dictionary,noise_model):
    weights_dictionary = benchmark_hamiltonian['weights_dictionary']
    state= list(results_dictionary.keys())[0]
    mitigated_marginals_dictionary={}
    for subset in weights_dictionary.keys():
        mitigated_marginals_dictionary[subset]=cn.mitigation.mitigate_marginal(marginal=subset,results_dictionary=results_dictionary,noise_model=noise_model)
    

        


def run_mitigation_benchmarks_for_hamiltonians(benchmark_hamiltonian_list, joint_mitigated_marginals_dictionary, noise_model):
    mitigated_energies_dictionary = {}
    for benchmark_hamiltonian, mitigated_marginals_dictionary in zip(benchmark_hamiltonian_list, joint_mitigated_marginals_dictionary):
        
        mitigated_energies_dictionary[mitigated_marginals_dictionary.key()] = calculate_mitigated_energy_for_hamiltonian(benchmark_hamiltonian=benchmark_hamiltonian,mitigated_marginals_dictionary=mitigated_marginals_dictionary) 
    return mitigated_energies_dictionary 
        






#divides indices of benchmark hamiltonians into traning and stest set
#outputs two lists: traning_set and test_ set with indices of respective hamiltonians
def create_traning_and_test_set(number_of_hamiltonians, traning_set_cardinality):
    traning_set=[]
    test_set =[]
    while (len(traning_set))<traning_set_cardinality:
        index=np.random.randint(number_of_hamiltonians)
        if index not in traning_set:
            traning_set.append(index)
    for i in range(number_of_hamiltonians):
        if i not in traning_set:
            test_set.append(i)
    return traning_set, test_set


# function calculates mean and median of mitigation benchmarks over a test/traning set
# input: benchmarks_results_mitigation - dictionary with mitigation results, traning_set - indicies of hamiltonians that form the traning set


def calculate_results_test_set(benchmark_results_mitigation, hamiltonian_set):
    # loop over differnt noise models (keys of benchmark_results_mitigation dictionary) and benchmark results for individual hamiltonians
    traning_median_mean = {}
    for noise_model, result in benchmark_results_mitigation.items():
        median, mean = calculate_median_mean(result['errors_list'], hamiltonian_set)
        traning_median_mean[noise_model] = {'median': median, 'mean': mean}
    return traning_median_mean


# helepr function that calculates median and mean for a single noise model
def calculate_median_mean(errors_dictionary, hamiltonian_set):
    median_list = []
    mean = 0.0
    for index in hamiltonian_set:
        error = errors_dictionary[index]
        mean = mean + error
        median_list.append(error)
    mean = mean / len(hamiltonian_set)
    return np.median(median_list), mean


#creates list of data to be plotted as a alpha vs benchmark result plot
#input clusters_dictionary (dictionary with clusters and details), and mitigation_data: list returned by calculate_results_test_set function
#outputs three lists 2d lists: plot_alpha_list - list of alpha (clustering parameter) for different localities
# plot_median_list - list with medians for different alphas/ localities
# plot_mean_list  - list with means different alphas/ localities
#the first index in the list encodes locality e.g. alpha_list[1] corresponds to list of alphas for clustering algorithm with locality =2
#corresponding values of medians for locality=2 are found in plot_median_list[1]
#this data is later used to create plots
def cerate_data_alpha_plot(clusters_dictionary, mitigation_data):
    parameters_dictionary = {}
    # loop over cluster assigments from clustering algorithm
    for cluster, clustering_params in clusters_dictionary.items():
        # this dictionary will store localities and corresponding alphas for a given cluster as well as median and mean for clustering
        locality_dic = {}
        max_locality = 0

        if clustering_params != None:
            # this is a loop over different configurations of localities and alpha for a given cluster structure
            for item in clustering_params:
                # data is rewritten as dictionary of a form locality : list of alphas
                if item[0] > max_locality:
                    max_locality = item[0]
                if item[0] in locality_dic.keys():
                    temp_list = locality_dic[item[0]]
                    temp_list.append(item[1])
                    locality_dic[item[0]] = temp_list

                else:

                    locality_dic[item[0]] = [item[1]]

        else:
            locality_dic[1] = [1]

        # to a given cluster value of meadian and mean of mitigation benchmark is added
        if 'median' in mitigation_data[cluster].keys():
            locality_dic['meadian'] = mitigation_data[cluster]['median']
            locality_dic['mean'] = mitigation_data[cluster]['mean']
        else:
            locality_dic['meadian'] = statistics.median(list(mitigation_data[cluster].values()))
            locality_dic['mean'] = statistics.mean(list(mitigation_data[cluster].values()))
        parameters_dictionary[cluster] = locality_dic
        # print(parameters_dictionary[cluster])

    plot_alpha_list = [[] for i in range(2, max_locality + 2)]
    plot_median_list = [[] for i in range(2, max_locality + 2)]
    plot_mean_list = [[] for i in range(2, max_locality + 2)]

    for clusters, parameters in parameters_dictionary.items():
        for i in range(1, max_locality + 1):
            if i in parameters.keys():
                for item in parameters[i]:
                    plot_alpha_list[i - 1].append(item)
                    plot_median_list[i - 1].append(parameters['meadian'])
                    plot_mean_list[i - 1].append(parameters['mean'])

    return plot_alpha_list, plot_median_list, plot_mean_list

#this function creates matplotlib plots for data created with cerate_data_alpha_plot
#plot_x_list is usually a list of alphas
#plot_y_list is a list of medians/means
#separable_value is a value for uncorelated noise model
#label is the plot description and name of the file where results are saved
def create_plots(plot_x_list, plot_y_list, separable_value, label,y_min=0,y_max=0.5):
    legend = []
    for i in range(len(plot_x_list)):

        # this if exludes plotting results for uncorelated noise model, they are added as a line below
        if len(plot_x_list[i]) > 1:
            plt.plot(plot_x_list[i], plot_y_list[i], marker='o', linestyle='none')
            title = label
            legend.append(f'Locality {i + 1}')
            min_val = min(plot_y_list[i])
            if min_val < separable_value:
                min_x = plot_x_list[i][plot_y_list[i].index(min_val)]
                plt.annotate(str(np.round(min_val, 5)), xy=(min_x, min_val))

            plt.title(title)
            fname = title + ".png"
    plt.ylim(y_min, y_max)

    legend.append("Uncorrelated")

    plt.plot([plot_x_list[1][0], plot_x_list[1][-1]], [separable_value, separable_value])
    plt.annotate(str(np.round(separable_value, 5)), xy=(2.5, separable_value))
    plt.legend(legend)
    plt.xlabel("alpha")
    plt.savefig(fname)


